// ============================================================================
// BDreaderGAMEBDreadLevelGAMEBD.cpp
// ----------------------------------------------------------------------------
// (c) 2022-2024 by HerzAusGold
//               Hubert Haberkorn
//               herz-aus-gold-42@gmx.de
//               http://www.arthurdent.de/
// ----------------------------------------------------------------------------
// The above copyright notice and this permission notice 
// shall be included in all copies or substantial portions 
// of the Software.
// ----------------------------------------------------------------------------
// read file dump from "BoulderDsh" (by Guido Mersmann) 
// and extract level data to -> ".gamebd" file
// generate "levelinfo.conf" with some extra info
// ============================================================================

#include "stdio.h"

#include <string>
#include <iostream>
#include <vector>
#include <filesystem>

#include "BDreaderGAMEBDmain.h"
#include "BDreaderGAMEBDutil.h"
#include "BDreaderGAMEBDoutput.h"
#include "BDreaderGAMEBDoutputCave.h"
#include "BDreaderGAMEBDoutputCaveGAMEBD.h"
#include "BDreaderGAMEBDoutputCaveBD64.h"
#include "BDreaderGAMEBDoutputCaveMagic64.h"
#include "BDreaderGAMEBDwriteLevelGAMEBD.h"


using namespace std;


/*---------------------------------------------------------------------------*/
static anchor_t            anchor {0};
static caveListHeader_t    caveListHeader {0};
static caveHeader_t        caveHeader {0};
static cave_t              cave {0};

static caveNew_t           caveNew {0};


/*---------------------------------------------------------------------------*/
static bool NextLevel(int showAllLevels) {
    static int alwaysweiter = 0;
    if (showAllLevels) {
        alwaysweiter = 1;
    }
    int weiter = 0;
    if (!alwaysweiter) {
        cout << "--> NextLevel (-1 always): ";
        cin >> weiter;
        if (weiter == -1) {
            alwaysweiter = 1;
        }
    } else {
        weiter = 1;
    }
    return (weiter != 0);
}


/*---- readLevelGAMEBD -----------------------------------------------------------*/
int readLevelGAMEBD(string directory, 
                    string gamename, 
                    string filename, 
                    string offsetStr, 
                    int i_levelNo,
                    int i_doFileWrite)
{
    int      levels     = 0;
    int      actLevelNo = 0;

    char textGameEditor[] = "Game Editor";

    size_t              rd = 0;
    byte_t              headerOffsetBytes[4];

    // init
    memset(&anchor,         0, sizeof(anchor));
    memset(&caveListHeader, 0, sizeof(caveListHeader));
    memset(&caveHeader,     0, sizeof(caveHeader));
    memset(&cave,           0, sizeof(cave));

    // show properties
    int           showFilename          = 0;  // 0;
    int           showOffset            = 0;
    int           showCaveListHeader    = 0;

    int           showCaveHeader        = 0;  // 1
    int           showCave              = 1;
    int           showCaveBytes         = 0;  // 1
    int           showCaveReplGAMEBD    = 1;
    int           showCaveReplBD64      = 0;
    int           showCaveReplMagic64   = 1;
    int           showAllLevels         = 1;

    int           cleanMemory           = 0;  // 0;  // 0 for compare

    int           doFileWrite           = i_doFileWrite;
    int           isError               = 0;

#ifdef VIEWER_ONLY
    showFilename            = 0;
    showOffset              = 0;
    showCaveListHeader      = 0;
    //--
    showCaveHeader          = 0;
    showCave                = 1;
    showCaveBytes           = 0;
    showCaveReplGAMEBD      = 1;
    showCaveReplBD64        = 0;
    showCaveReplMagic64     = 0;
    showAllLevels           = 0;
    cleanMemory             = 0;
    doFileWrite             = 0;
    if (i_levelNo == -1) {
        showAllLevels  = 1;
    }
#endif

    vector<caveNew_t>   caveNewArr;
    long                caveNewOffset = writeLevelGAMEBDgetCaveNewOffset();  // starting offset

    cout << endl;
    cout << "---===--- Boulderdaesh-Reader ---===---" << endl;
    if (showFilename) {
        printf("--- Filename: %s\n", filename.c_str());
    }
    if (showOffset) {
        printf("Size of caveheader      : dec: %d hex: %x block16: %lf \n", sizeof(caveHeader), sizeof(caveHeader), sizeof(caveHeader)/16.0);  //296  0x128  18 
        printf("Size of caveheader 0x128: %x\n", sizeof(caveHeader));
    }

    #pragma region "determine anchor"
    long anchorOffset = hex2dec(offsetStr);
    long anchorNull   = anchorOffset + 4;
    if (showOffset) {
        printf("--- Offset  : %s\n", offsetStr.c_str());
    }
    #pragma endregion

    cout << string(80, '-') << endl;
    int fok = 0;
    FILE *fp = fopen(filename.c_str(), "rb");
    if (fp != NULL) {
        //---- search anchor ----
        #pragma region "search anchor"

        if (showOffset) {
            cout << "search AnchorOffset: " << dec2hex(anchorOffset) << endl;
        }
        int anchorSkip = 0;
        while(1) {
            fok = fseek(fp, anchorOffset, SEEK_SET);
            if (fok == 0) {
                rd = fread(&anchor, sizeof(anchor_t), 1, fp);
                if (rd > 0) {
                    const char *cmpPtr = NULL;
                    cmpPtr = strstr((const char*) &anchor.gameEditorText[0], textGameEditor);
                    if (cmpPtr != (const char*) &anchor.gameEditorText[0]) {
                        anchorOffset += 2;
                        anchorSkip   += 2;
                    } else {
                        break;
                    }
                }
            }
        }
        anchorNull = anchorOffset + 4;

        if (showOffset) {
            cout << "--find Anchor(skip): " << anchorSkip << endl;
            cout << "--find AnchorOffset: " << dec2hex(anchorOffset) << endl;
        }
        if (showOffset) {
            outputAnchor(anchor);  // anchor
        }
        #pragma endregion

        //----  the levels ----
        cout << string(80, '-') << endl;
        memcpy(&headerOffsetBytes[0], &anchor.caveFirstLevelAddr[0], 4);

        while(1) {
            //---- determine caveListHeader offset (from list) ----
            #pragma region "determine caveListHeader"

            long headeroffset = bytes2offset(headerOffsetBytes);
            if (i_levelNo <= -1 || (actLevelNo+1) == i_levelNo) {
                cout << string(70, '=') << endl;
            }
            if (showOffset) {
                cout << "HeaderOffset: " << dec2hex(headeroffset) << endl;
            }
            #pragma endregion

            if (headeroffset == anchorNull) {
                break;  // anchorNull reached - no more Levels
            }
            actLevelNo++;
            

            //---- read caveListHeader ----
            #pragma region "read caveListHeader"
            fok = fseek(fp, headeroffset, SEEK_SET);
            if (fok == 0) {
                rd = fread(&caveListHeader, sizeof(caveListHeader_t), 1, fp);
                if (rd > 0) {
                    if (i_levelNo <= -1 || actLevelNo == i_levelNo) {
                        if (showCaveListHeader) {
                            cout << string(70, '-') << endl;
                            outputCaveListHeader(caveListHeader);
                        }
                    }
                }
            }
            #pragma endregion

            //---- copy caveNextLevelAddr to headerOffsetBytes ----
            memcpy(&headerOffsetBytes[0], &caveListHeader.caveNextLevelAddr[0], 4);

            //---- determine cave header (from start) ----
            #pragma region "determine caveHeader"

            long caveoffset = bytes2offset(&caveListHeader.caveStartAddr[0]);
            if (i_levelNo <= -1 || actLevelNo == i_levelNo) {
                if (showOffset) {
                    cout << "CaveOffset  : " << dec2hex(caveoffset) << endl;
                }
            }
            #pragma endregion

            //---- read caveHeader ----
            #pragma region "read caveHeader"

            fok = fseek(fp, caveoffset, SEEK_SET);
            if (fok == 0) {
                rd = fread(&caveHeader, sizeof(caveHeader_t), 1, fp);
                if (rd > 0) {
                    if (i_levelNo <= -1 || actLevelNo == i_levelNo) {
                        if (showCaveHeader) {
                            cout << string(70, '-') << endl;
                            string strCaveHeader = "caveheader";
                            string strCaveName = string((const char*)&caveHeader.caveName[0]);
                            strCaveHeader += string("        : ") + strCaveName;
                            if (showOffset) {
                                outbytes(strCaveHeader.c_str(), (byte_t*)&caveHeader, sizeof(caveHeader));
                            } else {
                                outbytes(strCaveHeader.c_str(), (byte_t*)&caveHeader, sizeof(caveHeader)-4);
                            }
                        }
                        outputCaveHeader(caveHeader, showOffset, showFilename,
                                         filename, gamename);
                    }
                }
            }
            #pragma endregion

            //---- determine cave ----
            #pragma region "determine cave"

            if (i_levelNo <= -1 || actLevelNo == i_levelNo) {
                if (showCave) {
                    printf ("~~~~~~~~~~~~~~~~ Cave: %s", &caveHeader.caveName[0]);
                    if (showFilename) {
                        printf("\t -- %s\n", filename.c_str());
                    } else {
                        printf("\t -- %s\n", gamename.c_str());
                    }
                }
            }
            long addrfrom      = bytes2offset(&caveListHeader.caveStartAddr[0]);
            long gameheaderlen = convBytes2offset(&caveHeader.headsize[0], sizeof(caveHeader.headsize));
            if (gameheaderlen != 0x0128) {
                handleError("-E- WRONG gameheaderlen !!");
                isError = true;
            }
            long addrlev  = addrfrom + 0x128;
            if (i_levelNo <= -1 || actLevelNo == i_levelNo) {
                 if (showCave && showOffset) {
                     cout << "AddrLev: " << dec2hex(addrlev)  << endl;
                 }
            }

            long cavelen  = convBytes2offset(&caveHeader.levelLength[0], 2);

            #pragma endregion

            //---- read cave ----
            #pragma region "read cave"

            fok = fseek(fp, addrlev, SEEK_SET);
            if (fok == 0) {
                long   index   = 0;
                while(1) {
                    byte_t levb = 0;
                    rd = fread(&levb, 1, 1, fp);
                    if (rd > 0) {
                        cave.cave[index] = levb;
                        index++;
                        if (levb == 0xFF) {
                            // read element
                            rd = fread(&levb, 1, 1, fp);
                            if (rd > 0) {
                                cave.cave[index] = levb;
                                index++;
                            }
                            if (levb == 0xFF) {
                                break;
                            }
                            // read numberOf
                            rd = fread(&levb, 1, 1, fp);
                            if (rd > 0) {
                                cave.cave[index] = levb;
                                index++;
                            }
                        }
                    }
                }
                // check length
                if (index != cavelen) {
                    cout << "Wrong cavelen!" << " index: " << index << " cavelen: " << cavelen << endl;
                    handleError("Wrong cavelen!");
                    isError = true;
                }
                // output
                if (i_levelNo <= -1 || actLevelNo == i_levelNo) {
                    if (showCave) {
                        int maxx = convBytes2offset(caveHeader.sizex, 2);
                        int maxy = convBytes2offset(caveHeader.sizey, 2);
                        if (showCaveBytes) {
                            outbytes("cave", (byte_t*)&cave, cavelen);
                        }
                        if (showCaveReplGAMEBD) {
                            outputCave(cave, maxx, maxy, cavelen,
                                       convertGAMEBD);
                        }
                        if (showCaveReplBD64) {
                            outputCave(cave, maxx, maxy, cavelen,
                                       convertBD64);
                        }
                        if (showCaveReplMagic64) {
                            outputCave(cave, maxx, maxy, cavelen,
                                       convertMagic64);
                        }
                    }
                }
            }

            #pragma endregion


            //---- caveAdd ----
            #pragma region "create caveNew and add to caveNewArr"

            memset(&caveNew, 0, sizeof(caveNew));
            caveNew.listHeaderLen = sizeof(caveListHeader_t);
            caveNew.caveHeaderLen = sizeof(caveHeader_t);
            caveNew.caveLen       = cavelen;
            caveNew.toLastByteLen = caveNew.listHeaderLen + caveNew.caveHeaderLen + caveNew.caveLen;
            caveNew.extraBytesLen = 2;
            if (caveNew.toLastByteLen % 2 == 1) {
                caveNew.extraBytesLen += 1;
            }
            caveNew.caveNewOffset = caveNewOffset;
            memcpy(&caveNew.unknownBytes01[0], caveListHeader.unknownBytes01, sizeof(caveNew.unknownBytes01));

            // header
            memcpy(&caveNew.caveheader, &caveHeader, sizeof(caveHeader_t));
            memset(&caveNew.caveheader.levelScoreStartAddr[0], 0, sizeof(caveNew.caveheader.levelScoreStartAddr));  // always
            if (cleanMemory) {
                // clean caveHeader
                memset(&caveNew.caveheader.caveName[0], 0, sizeof(caveNew.caveheader.caveName));
                memset(&caveNew.caveheader.gfxName[0], 0, sizeof(caveNew.caveheader.gfxName));
                memset(&caveNew.caveheader.fntName[0], 0, sizeof(caveNew.caveheader.fntName));
                memset(&caveNew.caveheader.typeName[0], 0, sizeof(caveNew.caveheader.typeName));
                memset(&caveNew.caveheader.musName[0], 0, sizeof(caveNew.caveheader.musName));
                memset(&caveNew.caveheader.colorSetName[0], 0, sizeof(caveNew.caveheader.colorSetName));
                // set caveHeader
                strncpy((char*)&caveNew.caveheader.caveName[0], (char*)&caveHeader.caveName[0], sizeof(caveNew.caveheader.caveName));
                strncpy((char*)&caveNew.caveheader.gfxName[0], (char*)&caveHeader.gfxName[0], sizeof(caveNew.caveheader.gfxName));
                strncpy((char*)&caveNew.caveheader.fntName[0], (char*)&caveHeader.fntName[0], sizeof(caveNew.caveheader.fntName));
                strncpy((char*)&caveNew.caveheader.typeName[0], (char*)&caveHeader.typeName[0], sizeof(caveNew.caveheader.typeName));
                strncpy((char*)&caveNew.caveheader.musName[0], (char*)&caveHeader.musName[0], sizeof(caveNew.caveheader.musName));
                strncpy((char*)&caveNew.caveheader.colorSetName[0], (char*)&caveHeader.colorSetName[0], sizeof(caveNew.caveheader.colorSetName));
            }

            // cave
            memset(&caveNew.cave, 0, sizeof(caveNew.cave));
            memcpy(&caveNew.cave, &cave, cavelen);
            // extraBytes
            memset(&caveNew.extraBytes, 0, sizeof(caveNew.extraBytes));

            caveNewArr.push_back(caveNew);
            #pragma endregion

            levels += 1;

            //---- show caveNew (only offset) ----
            #pragma region "show caveNew"
            if (i_levelNo <= -1 || actLevelNo == i_levelNo) {
                if (showOffset) {
                    cout << "headeroffset: "   << tohexstr(headeroffset) 
                         << " listheader: "    << caveNew.listHeaderLen 
                         << " caveheader: "    << caveNew.caveHeaderLen
                         << " cavelen: "       << caveNew.caveLen 
                         << " toLastByteLen: " << caveNew.toLastByteLen 
                         << " extraBytesLen: " << caveNew.extraBytesLen 
                         << " caveNewOffset: " << caveNew.caveNewOffset 
                         << endl;
                }
            }
            #pragma endregion

            caveNewOffset += caveNew.toLastByteLen + caveNew.extraBytesLen; 

            if (i_levelNo == -2) {
                if (!NextLevel(showAllLevels)) {
                    break;
                }
            }
        }
        fclose(fp);
    }

    {   //---- Write ----
        bool fileWrite = doFileWrite;
        if (fileWrite && !isError) {
            writeLevelGAMEBD(directory, gamename, filename,
                             cleanMemory, anchor, 
                             caveNewArr);
        }
        
    }
    return levels;
}


/*---- doReadLevelGAMEBD -----------------------------------------------------------------*/
int doReadLevelGAMEBD (string filename, string offsetStr, int i_levelNo, int i_doFileWrite)
{ 
    string fname = filename;
    string directory = "";
    string subdir    = "";
    string prefixdir = "";

    string gamename  = "";

    size_t last_idx  = 0;
    
    last_idx  = filename.rfind('\\');
    if (std::string::npos != last_idx) {
        directory = filename.substr(0, last_idx);
        gamename  = filename.substr(last_idx+1);
    } else {
        directory = "";
        gamename  = filename;
    }

    last_idx  = gamename.rfind(".game");
    if (std::string::npos != last_idx) {
        gamename  = gamename.substr(0, last_idx);
    }

    last_idx  = directory.rfind('\\');
    if (std::string::npos != last_idx) {
        subdir    = directory.substr(last_idx+1);
        prefixdir = directory.substr(0,last_idx);
    } else {
        subdir    = "";
        prefixdir = directory;
    }

    last_idx  = prefixdir.rfind('\\');
    if (std::string::npos != last_idx) {
        prefixdir = prefixdir.substr(0,last_idx);
    }

    string newPath   = "";
    if (i_doFileWrite) {
        if (prefixdir == "") {
            prefixdir = "C:\\temp";
        }
        newPath = prefixdir + "\\GAMEBD\\" ;
        filesystem::path dir1{newPath};
        filesystem::create_directory(dir1);

        if (subdir != "") {
            newPath = prefixdir + "\\GAMEBD\\" + subdir ;
            filesystem::path dir2{newPath};
            filesystem::create_directory(dir2);

            if (gamename != "") {
                newPath = prefixdir + "\\GAMEBD\\" + subdir + "\\" + gamename;
                filesystem::path dir3{newPath};
                filesystem::create_directory(dir3);
            }
        } else {
            // subdir blank
            if (gamename != "") {
                newPath = prefixdir + "\\GAMEBD\\" + gamename;
                filesystem::path dir3{newPath};
                filesystem::create_directory(dir3);
            }
        }
    }

    // readLevel
    int  fileLevels = readLevelGAMEBD(newPath, gamename, filename, offsetStr, i_levelNo, i_doFileWrite);

    // write levelInfo.conf
    char levelsStr[10];
    sprintf(levelsStr, "%d", fileLevels);

    if (i_doFileWrite) {
        FILE *fp = NULL;
        string levelInfoFileName = newPath + "\\" + "levelinfo.conf";
        fp = fopen(levelInfoFileName.c_str(), "w");
        if (fp != NULL) {
            string buf;
            buf = "file_identifier:                ROCKSNDIAMONDS_LEVELINFO_FILE_VERSION_1.2\n";
            fputs(buf.c_str(), fp);
            buf = "\n";
            fputs(buf.c_str(), fp);
            buf = "name:                           " + gamename + "\n";
            fputs(buf.c_str(), fp);
            buf = "author:                         " + subdir + "\n";
            fputs(buf.c_str(), fp);
            buf = "\n";
            fputs(buf.c_str(), fp);
            buf = "levels:                         " + string(levelsStr) + "\n";
            fputs(buf.c_str(), fp);
            buf = "first_level:                    " + string("1") + "\n";
            fputs(buf.c_str(), fp);
            buf = "readonly:                       " + string("true") + "\n";
            fputs(buf.c_str(), fp);
            buf = "\n";
            fputs(buf.c_str(), fp);
            fclose(fp);
        }
    }

    if (subdir != "") {
        newPath = prefixdir + "\\GAMEBD\\" + subdir;
    } else {
        newPath = prefixdir + "\\GAMEBD";
    }
    if (i_doFileWrite) {

        FILE *fp = NULL;
        string levelInfoFileName = newPath + "\\" + "levelinfo.conf";
        fp = fopen(levelInfoFileName.c_str(), "w");
        if (fp != NULL) {
            string buf;
            buf = "file_identifier:                ROCKSNDIAMONDS_LEVELINFO_FILE_VERSION_1.2\n";
            fputs(buf.c_str(), fp);
            buf = "\n";
            fputs(buf.c_str(), fp);
            buf = "name:                           " + subdir + "\n";
            fputs(buf.c_str(), fp);
            buf = "author:                         " + string("Various Authors") + "\n";
            fputs(buf.c_str(), fp);
            buf = "\n";
            fputs(buf.c_str(), fp);
            buf = "sort_priority:                  " + string("200") + "\n";
            fputs(buf.c_str(), fp);
            buf = "level_group:                    " + string("true") + "\n";
            fputs(buf.c_str(), fp);
            buf = "\n";
            fputs(buf.c_str(), fp);
            fclose(fp);
        }
    }
    return fileLevels;
}
