// ============================================================================
// BDreaderGAMEBDutil.cpp
// ----------------------------------------------------------------------------
// (c) 2022-2024 by HerzAusGold
//               Hubert Haberkorn
//               herz-aus-gold-42@gmx.de
//               http://www.arthurdent.de/
// ----------------------------------------------------------------------------
// The above copyright notice and this permission notice 
// shall be included in all copies or substantial portions 
// of the Software.
// ----------------------------------------------------------------------------
// read file dump from "BoulderDsh" (by Guido Mersmann) 
// and extract level data to -> ".gamebd" file
// generate "levelinfo.conf" with some extra info
// ============================================================================

#include "BDreaderGAMEBDutil.h"
#include <string>
#include <iostream>
#include <iomanip>
#include <sstream>

static char hexarr[] = "0123456789ABCDEF";

//---------------------------------------------------------------------------
// handleError (for debugging)
//---------------------------------------------------------------------------
void handleError(const char * errMsg) {
    int weiter = 0;
    std::cout << "+++++++++++++ " << errMsg << std::endl;
    std::cout << "++ weiter: ";
    std::cin >> weiter;
}


//---------------------------------------------------------------------------
// (static) decimalToHex -- convert long to hex
//---------------------------------------------------------------------------
static std::string decimalToHex(long dec)
{
    std::string result = "";
    long rem = 0; 
    while (dec > 0) {
        rem = dec % 16;
        result = hexarr[rem] + result;
        dec = dec / 16;
    }
    return result;
}

//---------------------------------------------------------------------------
// dec2hex -- convert long to hex (fill with zero)
//---------------------------------------------------------------------------
std::string dec2hex(long dec)
{
    std::string result = decimalToHex(dec);
    if (strlen(result.c_str()) == 0) {
        result = "00";
    } else if (strlen(result.c_str()) % 2 == 1) {
        result = "0" + result;
    }
    return result;
}

//---------------------------------------------------------------------------
// hex2dec -- convert hex to long
//---------------------------------------------------------------------------
long hex2dec(std::string hex_value)
{
    long decimal_value {};
    std::stringstream ss;
    ss  << hex_value ; // std::string hex_value
    ss >> std::hex >> decimal_value ; //long decimal_value
    return decimal_value;
}

//---------------------------------------------------------------------------
// outbytes -- print title -- then hex values (maxlen is parameter)
// bundlesize is 1 -- blocksize is 16
//---------------------------------------------------------------------------
void outbytes (const char *title, byte_t *what, size_t len,
               size_t bundlesize, size_t blocksize) 
{
    if (title != NULL) {
        printf("== %s\n", title);
    }
    size_t cnt = 0;
    for (size_t i=0; i<len; i++) {
        printf("%s", dec2hex(what[i]).c_str());
        cnt++;
        if (cnt % bundlesize == 0) {
            printf(" ");
        }
        if (cnt >= blocksize) {
            cnt = 0;
            printf("\n");
        }
    }
    printf("\n");
}

//---------------------------------------------------------------------------
// outtext -- print title -- then text (maxlen is parameter)
//---------------------------------------------------------------------------
void outtext (const char *title, byte_t *what, size_t len) 
{
    if (title != NULL) {
        printf("== %s: ", title);
    }
    for (size_t i=0; i<len; i++) {
        if (what[i] == 0) {
            break;
        }
        printf("%c", what[i]);
    }
    printf("\n");
}

//---------------------------------------------------------------------------
// convBytes2offset -- convert bytes to long (size is parameter)
//---------------------------------------------------------------------------
long convBytes2offset (byte_t *what, size_t len) 
{
    long result = 0;
    long multi  = 1;
    for (int i=len-1; i>=0; i--) {
        result += what[i] * multi;
        multi *= 256;
    }
    return result;
}

//---------------------------------------------------------------------------
// bytes2offset -- convert bytes to long (default size is 4)
//---------------------------------------------------------------------------
long bytes2offset (byte_t *what) 
{
    return convBytes2offset(what, 4);
}

//---------------------------------------------------------------------------
// tostr -- convert long to string - with "_ltoa" - basis 10
//---------------------------------------------------------------------------
std::string tostr(long w)
{
    char hstr[20] = {0};
    _ltoa(w, hstr, 10);
    return std::string(hstr);
}

//---------------------------------------------------------------------------
// tohexstr -- convert long to string - with "_ltoa" - basis 16
// fill with zeros (default length 4)
//---------------------------------------------------------------------------
std::string tohexstr(long w, size_t sz)
{
    char hstra[20] = {0};
    char hstrb[20] = {0};
    _ltoa(w, hstra, 16);
    while (strlen(hstra) < sz*2) {
      sprintf(hstrb, "0%s", hstra);
      strcpy(hstra, hstrb);
    }
    return std::string(hstra);
}

//---------------------------------------------------------------------------
// HexToBytes -- convert hexstr to bytes (sz is size of bytes)
//---------------------------------------------------------------------------
void HexToBytes(byte_t *bytes, const char* hexstr, size_t sz)
{
   char hstr[4] = {0};
   memset (bytes, 0, sz);
   for (size_t i=0; i<sz; i++) {
        strncpy(hstr, &hexstr[i*2], 2);
        hstr[2] = 0;
        bytes[i] = (byte_t)hex2dec(hstr);
   }
}

//---------------------------------------------------------------------------
// DecToBytes -- convert long to bytes (sz is size of bytes)
//---------------------------------------------------------------------------
void DecToBytes(byte_t *bytes, long deci, size_t sz)
{
   std::string hstr= tohexstr(deci, sz);
   HexToBytes(bytes, hstr.c_str(), sz);
}
