// ============================================================================
// BDreaderGAMEBDwriteLevelGAMEBD.cpp
// ----------------------------------------------------------------------------
// (c) 2022-2024 by HerzAusGold
//               Hubert Haberkorn
//               herz-aus-gold-42@gmx.de
//               http://www.arthurdent.de/
// ----------------------------------------------------------------------------
// The above copyright notice and this permission notice 
// shall be included in all copies or substantial portions 
// of the Software.
// ----------------------------------------------------------------------------
// read file dump from "BoulderDsh" (by Guido Mersmann) 
// and extract level data to -> ".gamebd" file
// generate "levelinfo.conf" with some extra info
// ============================================================================

#include "stdio.h"

#include <string>
#include <iostream>
#include <vector>
#include <filesystem>

#include "BDreaderGAMEBDmain.h"
#include "BDreaderGAMEBDutil.h"

using namespace std;

static char     m_magic[]  = "GAMEBD\0\0";
static long     m_magiclen = sizeof(m_magic) - 1;      // magiclen = 8 -> because sizeof(magic) is 9 (8 bytes in string + termination)
static long     m_startHeaderOffset = 8;               // numberOfLevels(4) + caveCurrent(4)
static long     m_caveNewOffset = m_magiclen + m_startHeaderOffset + sizeof(anchor_t);

/*---- writeLevelGAMEBDgetCaveNewOffset -------------------------------------------*/
long writeLevelGAMEBDgetCaveNewOffset()
{
    return m_caveNewOffset;
}

/*---- writeLevelGAMEBD -----------------------------------------------------------*/
void writeLevelGAMEBD(
        string              directory, 
        string              gamename, 
        string              filename,
        int                 cleanMemory,
        anchor_t            &anchor,
        vector<caveNew_t>   &caveNewArr
)
{

    size_t wrc;
    string m_filename = directory + "\\" + gamename + ".gamebd"; // filename + "bd";  // "gamebd"
    FILE *fp = fopen(m_filename.c_str(), "wb");
    if (fp != NULL) {
        long     startHeader = 0;
        wrc = fwrite((const void*)&m_magic, m_magiclen, 1, fp);     // only magic
        if (wrc != 1) {
            handleError("Write Error");
        }
        startHeader = m_magiclen + m_startHeaderOffset;

        // header
        // byte_t      caveFirstLevelAddr[4];
        // byte_t      caveNullLevelAddr[4];
        // byte_t      caveLastLevelAddr[4];
        // byte_t      editorPossible[4];        // 0100 dez // editor possible 0000
        // byte_t      gameEditorText[6*16+4];   
        // byte_t      gameFileName[132];
        long m_caveFirstLevelAddr = startHeader + sizeof(anchor_t);
        long m_caveNullLevelAddr  = 0;
        long m_caveLastLevelAddr  = 0;

        caveNew_t & cavefirst = caveNewArr[0];
        m_caveFirstLevelAddr  = cavefirst.caveNewOffset;
        caveNew_t & cavelast  = caveNewArr[caveNewArr.size()-1];
        m_caveLastLevelAddr   = cavelast.caveNewOffset;

        // startheader numberOfLevels
        byte_t numberOfLevels[4];
        DecToBytes(numberOfLevels, caveNewArr.size(), 4);  
        wrc = fwrite((const void*)&numberOfLevels, sizeof(numberOfLevels), 1, fp);
        if (wrc != 1) {
            handleError("Write Error");
        }

        // startheader caveCurrent
        byte_t caveCurrent[4];
        DecToBytes(caveCurrent, m_caveFirstLevelAddr, 4);  
        wrc = fwrite((const void*)&caveCurrent, sizeof(caveCurrent), 1, fp);
        if (wrc != 1) {
            handleError("Write Error");
        }

        // build anchorNew
        anchor_t anchorNew ;
        memset(&anchorNew, 0, sizeof(anchorNew));
        DecToBytes(anchorNew.caveFirstLevelAddr, m_caveFirstLevelAddr, 4);  
        DecToBytes(anchorNew.caveNullLevelAddr,  m_caveNullLevelAddr, 4);  
        DecToBytes(anchorNew.caveLastLevelAddr,  m_caveLastLevelAddr, 4);  
        memcpy(anchorNew.editorPossible, anchor.editorPossible, sizeof(anchor.editorPossible));
        if (cleanMemory) {
            memset(anchorNew.gameEditorText, 0, sizeof(anchorNew.gameEditorText));
            memset(anchorNew.gameFileName, 0, sizeof(anchorNew.gameFileName));
        }
        strncpy((char*)anchorNew.gameEditorText, (char*)anchor.gameEditorText, sizeof(anchor.gameEditorText));
        strncpy((char*)anchorNew.gameFileName,   (char*)anchor.gameFileName,   sizeof(anchor.gameFileName));

        // write anchorNew
        wrc = fwrite((const void*)&anchorNew, sizeof(anchor_t), 1, fp);
        if (wrc != 1) {
            handleError("Write Error");
        }

        // write caves
        for (size_t i=0; i < caveNewArr.size(); i++) {
            caveNew_t & caveCurr = caveNewArr[i];
            if (i == 0) {
                caveCurr.caveLastLevelAddr = startHeader+4;
            } else {
                caveNew_t & caveLast = caveNewArr[i-1];
                caveCurr.caveLastLevelAddr = caveLast.caveNewOffset;
            }
            if (i == caveNewArr.size()-1) {
                caveCurr.caveNextLevelAddr = startHeader+4; 
            } else {
                caveCurr.caveNextLevelAddr = caveCurr.caveNewOffset + caveCurr.toLastByteLen + caveCurr.extraBytesLen; 
            }
            caveCurr.caveStartAddr = caveCurr.caveNewOffset + 4 + 4 + 2 + 4;  // caveLastLevelAddr + caveNextLevelAddr + unkownBytes + caveStartLevelAddr

            // build caveListHeaderNew
            caveListHeader_t caveListHeaderNew;
            memset(&caveListHeaderNew, 0, sizeof(caveListHeaderNew));
            DecToBytes(caveListHeaderNew.caveNextLevelAddr, caveCurr.caveNextLevelAddr, 4);
            DecToBytes(caveListHeaderNew.caveLastLevelAddr, caveCurr.caveLastLevelAddr, 4);
            DecToBytes(caveListHeaderNew.caveStartAddr, caveCurr.caveStartAddr, 4);
            memcpy(caveListHeaderNew.unknownBytes01, caveCurr.unknownBytes01, sizeof(caveListHeaderNew.unknownBytes01));

            // write caveListHeaderNew
            wrc = fwrite((const void*)&caveListHeaderNew, sizeof(caveListHeader_t), 1, fp);
            if (wrc != 1) {
                handleError("Write Error");
            }
            // write caveHeader
            wrc = fwrite((const void*)&caveCurr.caveheader, sizeof(caveCurr.caveheader), 1, fp);
            if (wrc != 1) {
                handleError("Write Error");
            }
            // write cave
            wrc = fwrite((const void*)&caveCurr.cave, caveCurr.caveLen, 1, fp);
            if (wrc != 1) {
                handleError("Write Error");
            }
            // write extraBytes
            if (caveCurr.extraBytesLen > 0) {
                wrc = fwrite((const void*)&caveCurr.extraBytes, caveCurr.extraBytesLen, 1, fp);
                if (wrc != 1) {
                    handleError("Write Error");
                }
            }
        }
        fclose(fp);
    }
}
