// ============================================================================
// BDreaderNIBBLEmain.cpp
// ----------------------------------------------------------------------------
// (c) 2022-2024 by HerzAusGold
//               Hubert Haberkorn
//               herz-aus-gold-42@gmx.de
//               http://www.arthurdent.de/
// ----------------------------------------------------------------------------
// The above copyright notice and this permission notice 
// shall be included in all copies or substantial portions 
// of the Software.
// ----------------------------------------------------------------------------
// read "raw" file dump and extract CAV and INT 
// generate "levelinfo.conf" with some extra info
// append "BDreaderNibble.csv" for overview at folder -> C:\\temp
// ============================================================================

#include "stdio.h"
#include <iostream>
#include <string>
#include <filesystem>

using namespace std;


#define BD_CAVE_WIDTH            40
#define BD_CAVE_HEIGHT           22
#define BD_INTERMISSION_WIDTH    20
#define BD_INTERMISSION_HEIGHT   12


typedef struct bd_bd64_data_s {
  char      fileid[2];
  char      ldata[440];
  char      cavedelay;
  char      unused01;
  char      cavetime;
  char      unused02;
  char      diamneeded;
  char      unused03;
  char      diamvalue;
  char      unused04;
  char      diambonus;
  char      unused05;
  char      slimperm;
  char      unused06;
  char      amoebatime;
  char      unused07;
  char      magwalltime;
  char      unused08;
  char      unknown1[18];
  char      flags;
  char      color1;
  char      unused09;
  char      color2;
  char      unused10;
  char      color3;
  char      unused11;
  char      color4;
  char      unused12;
  char      color5;
  char      unknown2[12];
} bd_bd64_data_t;


static unsigned char bd_bd64_fileid_head[2]          = { 0x00, 0xBE };
static unsigned char bd_bd64_flagid_intermission1[1] = { 0x01 }; // { 0x01 };
static unsigned char bd_bd64_flagid_intermission2[1] = { 0x3C }; // { 0x3C };
static unsigned char bd_bd64_cavedata[4]             = { 0x44, 0x44, 0x44, 0x44 }; // cave should have a wall
static unsigned char bd_bd64_cavenone[4]             = { 0x00, 0x00, 0x00, 0x00 }; // cave empty

static char prog[0x7000+1];
static bd_bd64_data_t bd_bd64_data;


/*--------------------------------------------------*/
static int chint (char ch) {
  int ia;
  ia = ch;
  if (ia < 0) {
    ia += 256;
  }
  return (ia);
}

/*--------------------------------------------------*/
static int checkborder(char *ldata, int lev)
{
    int  cnt1   = 0;
    int  cnt2   = 0;
    for (int i=0; i<BD_CAVE_WIDTH/2; i++) {       // erste Zeile
        if (ldata[i] == 0x44) {
           cnt1++;
        }
    }
    for (int i=BD_CAVE_WIDTH/2; i<BD_CAVE_WIDTH; i++) {       // zweite Zeile
        if (ldata[i] == 0x44) {
           cnt2++;
        }
    }
    int cnt =(cnt2 > cnt1)?cnt2:cnt1;
    return cnt;
} // checkborder

/*--------------------------------------------------*/
static int getLevMaxEasy()
{
    int cnt  = 0;
    int addr;
    int zaddr;
    // checkjoy
    unsigned char checkjoy[] = {0x20, 0x52, 0x41, 0xC9, 0x0F, 0xF0, 0x4A};  // JSR $4152; CMP #$0F; BEQ $547D
    // check prog
    addr = 0x542C;
    if (memcmp(&prog[addr], checkjoy, sizeof(checkjoy)) == 0) {
        addr = 0x543A;
        int offcave  = prog[addr];     // offset cave 0x0D or 0x01
        addr = 0x5444;
        int offtype  = prog[addr];     // offset type 0x0C or 0x00
        addr  = 0x5E8B; 
        zaddr = addr;
        for(int m=0;;m++){
            zaddr = addr + offtype;
            int levtype = prog[zaddr];
            if (levtype == 0x19 || levtype == 0x0E) {
                cnt++;
            } else {
                break;
            }
            addr = addr + offcave;
        }
    } 
    return cnt;
} // getLevMaxEasy

//--- main ---------------------------------
int main(int argc, char *argv[])
{
    bool   withoutProg = false;
    string defaultDir  = "C:\\temp\\BoulderDashTosec\\_d64\\";      //@@ Directory where the ".raw" files exists
    string filename;
    size_t last_idx;
    if (argc > 1) {
        filename = string(argv[1]);
    } else {
        char cfilename[1000] = {0};
        cerr << "Filename          : ";
        cin.getline(cfilename, sizeof(cfilename), '\n');
        filename = string(cfilename);
        // 
        last_idx   = filename.rfind(".raw");
        if (std::string::npos == last_idx) {
            filename = filename + ".raw";
        }
        last_idx   = filename.find(defaultDir);
        if (std::string::npos == last_idx) {
            filename = defaultDir + filename;
        }

    }
    int levMax = 0;
    int levMaxUser = 0;
    if (argc > 2) {
        levMaxUser = atoi(argv[2]);
    }

    string prefixname = "";
    last_idx   = filename.rfind('.');
    if (std::string::npos != last_idx) {
        prefixname = filename.substr(0,last_idx);
    } else {
        prefixname = filename;
    }
    string mainname = "";
    last_idx   = prefixname.rfind('\\');
    if (std::string::npos != last_idx) {
        mainname = prefixname.substr(last_idx+1);
    } else {
        mainname = prefixname;
    }
    string mainnameshort;
    string year;
    string author;
    string country;
    string other;
    int    bcnt = 0;
    bool   addch = true;
    for(size_t i=0; i<mainname.size(); i++){
        if (mainname[i] == '(') {
            addch = false;  
            bcnt++;
        }
        if (addch) {
            if (mainname[i] != '#') {   // comment
                mainnameshort += mainname[i];
            }
        } else {
            if (!(mainname[i] == '(' || mainname[i] == ')')) {
                switch (bcnt) {
                    case 1 : year    += mainname[i]; break;
                    case 2 : author  += mainname[i]; break;
                    case 3 : country += mainname[i]; break;
                    default: other   += mainname[i]; break;
                }
            }
        }
        if (mainname[i] == ')') {
            addch = true;  
        }
    }

    FILE* fp = fopen(filename.c_str(), "rb");
    if (fp != NULL) {
        // programm lesen
        int  progoffset = 0x7000;
        int  sz = 0;
        fseek(fp, 0L, SEEK_END);
        sz = ftell(fp);
        fseek(fp, 0L, SEEK_SET);
        if (sz < 0xD000) {
            withoutProg = true; 
            progoffset  = 0x02;  // header -- 2 Bytes lesen        
        }

        sz = fread(prog, 1, progoffset, fp);
        // check easy struct
        int levMaxEasy  = 0;
        if (!withoutProg) {
            levMaxEasy = getLevMaxEasy();
        }
        // caves lesen
        int  cavemem    = 0x7000;
        int  caveoffset = 0x0200;
        char cave[0x0200+1];
        int  cz;
        char levname[256];
        for(int lev=1; ; lev++)
        {
            memset(cave, 0, sizeof(cave));
            // read
            cz = fread(cave, 1, caveoffset, fp);
            // in die struktur kopieren
            char *ldataptr = &bd_bd64_data.ldata[0];
            memcpy(ldataptr, cave, sizeof(bd_bd64_data)-2);  // ldata

            // schreiben
            if (levMaxUser > 0 ||
                checkborder(&bd_bd64_data.ldata[0], lev) >= 6)  // BD_CAVE_WIDTH -> zwei Zeilen (2*20)
            {   
                int  x,y;
                int  maxx,maxy;
                bool intermission;
                // real level with wall
                cout << "Level    : " << lev << endl;
                cout << "CaveMem  : " << hex << cavemem << dec << endl;
                cavemem += caveoffset;
                //
                levMax = lev;

                /*--------------------------------------------------------*/
                /*--- intermission check ---------------------------------*/
                /*--------------------------------------------------------*/
                maxx = BD_CAVE_WIDTH;
                maxy = BD_CAVE_HEIGHT;
                /*- check1 -*/
                intermission = false;
                if ((memcmp(&bd_bd64_data.flags, bd_bd64_flagid_intermission1, 1) == 0) ||
                    (memcmp(&bd_bd64_data.flags, bd_bd64_flagid_intermission2, 1) == 0)) {
                    cout << "--- Intermission --(1)-\n";
                    intermission = true;
                } else {
                    cout << "--- Cave ----------(1)-\n";
                    intermission = false;
                }

                /*- check2 -*/
                intermission = true;
                for (y=0; y<maxy; y++) {
                    for (x=BD_INTERMISSION_WIDTH/2; x<maxx/2; x++) {
                        int ch  = chint (bd_bd64_data.ldata[y*maxx/2+x]);
                        if (ch != 0x44) {
                            intermission = false;
                            y = maxy+1;
                            break;
                        }      
                    }
                }
                string ext = ".cav";
                if (intermission) {
                    cout << "--- Intermission --(2)-\n";
                    ext = ".int";
                } else {
                    cout << "------ Cave -------(2)-\n";
                    ext = ".cav";
                }


                if (prefixname == filename) {
                    filesystem::path dir1{prefixname + "Dir"};
                    filesystem::create_directory(dir1);
                    sprintf(levname, "%s\\%2.2d%s", (prefixname + "Dir").c_str(), lev, ext.c_str());
                    cout << "Levelname: " << levname << endl;
                } else {
                    filesystem::path dir1{prefixname};
                    filesystem::create_directory(dir1);
                    sprintf(levname, "%s\\%2.2d%s", prefixname.c_str(), lev, ext.c_str());
                    cout << "Levelname: " << levname << endl;
                }

                FILE* fpw = fopen(levname, "wb");
                if (fpw != NULL) {
                    memcpy(bd_bd64_data.fileid, bd_bd64_fileid_head, sizeof(bd_bd64_data.fileid));
                    // header    
                    fwrite(bd_bd64_data.fileid, 1, sizeof(bd_bd64_data.fileid), fpw);
                    // cave
                    int sx = sizeof(bd_bd64_data);
                    fwrite(cave, 1, caveoffset - 0x10, fpw);  // 0x10 -- letzen 16 bytes nicht schreiben
                    fclose(fpw);
                }
                if (levMaxUser > 0 && levMax > levMaxUser) {
                    break;
                }
            } else {
                break;
            }
        }
        fclose(fp);

        if (levMax > 0)
        {
            sprintf(levname, "%s\\%s", prefixname.c_str(), "levelinfo.conf");
            FILE* fpw = fopen(levname, "w");
            string buf;

            cout << "--- Write levelinfo.conf ->" 
                 << " levMax: "  << std::to_string(levMax) 
                 << " levEasy: " << std::to_string(levMaxEasy) 
                 << " name: " << mainnameshort << endl;

            if (fpw != NULL) {
                buf = "file_identifier:                ROCKSNDIAMONDS_LEVELINFO_FILE_VERSION_1.2\n";
                fputs(buf.c_str(), fpw);
                buf = "\n";
                fputs(buf.c_str(), fpw);
                buf = "name:                           " + mainnameshort + "\n"; 
                fputs(buf.c_str(), fpw);
                buf = "author:                         " + author + "\n"; 
                fputs(buf.c_str(), fpw);
                buf = "\n";
                fputs(buf.c_str(), fpw);
                buf = "levels:                         " + std::to_string(levMax) + "\n";
                fputs(buf.c_str(), fpw);
                buf = "first_level:                    " + string("1") + "\n";
                fputs(buf.c_str(), fpw);
                buf = "readonly:                       " + string("true") + "\n"; 
                fputs(buf.c_str(), fpw);
                buf = "\n";
                fputs(buf.c_str(), fpw);

                buf = "#mainname:                      " + mainname + "\n"; 
                fputs(buf.c_str(), fpw);
                buf = "#year:                          " + year + "\n"; 
                fputs(buf.c_str(), fpw);
                buf = "#country:                       " + country + "\n"; 
                fputs(buf.c_str(), fpw);
                buf = "#other:                         " + other + "\n"; 
                fputs(buf.c_str(), fpw);
                buf = "#levMaxSelect:                  " + std::to_string(levMaxEasy) + "\n"; 
                fputs(buf.c_str(), fpw);

                buf = "\n";
                fputs(buf.c_str(), fpw);
                fclose(fpw);
            }
        }
        else
        {
            cout << "--- NO LEVELS found: " << " name: " << mainnameshort << " author: " << author << endl;
        }

        // write/append CSV
        if (levMax >= 0)  // immer schreiben
        {
            FILE* fpw = fopen("C:\\temp\\BDreaderNibble.csv", "a+");        //@@ CSV file 
            string buf;
            buf = mainname + ";" + 
                  mainnameshort + ";" + 
                  ";" +   // remark
                  ";" +   // check
                  std::to_string(levMax) + ";" + 
                  std::to_string(levMaxUser) + ";" + 
                  std::to_string(levMaxEasy) + ";" +
                  author  + ";" + 
                  year    + ";" +
                  country + ";" +
                  other   + ";" +
                  "\n";

            if (fpw != NULL) {
                fputs(buf.c_str(), fpw);
                fclose(fpw);
            }
        }
    }

}

