// ============================================================================
// BDreaderWLCprogMain.cpp
// ----------------------------------------------------------------------------
// (c) 2022-2024 by HerzAusGold
//               Hubert Haberkorn
//               herz-aus-gold-42@gmx.de
//               http://www.arthurdent.de/
// ----------------------------------------------------------------------------
// The above copyright notice and this permission notice 
// shall be included in all copies or substantial portions 
// of the Software.
// ----------------------------------------------------------------------------
// read "raw" file dump and extract the data with parameters.
// create .WLC file - level command data is stored 1:1
// ============================================================================

#include "stdio.h"
#include <iostream>
#include <string>
#include <vector>
#include <filesystem>
#include "decodeCave.h"
#include "decodeCaveCommon.h"
#include "decodeCaveBD02.h"

using namespace std;

unsigned char prog [0xFFFF+2]  = {0};
unsigned char cave [880+100]   = {0};
unsigned char para [0x1FF+10]  = {0};

unsigned char lowofftab   = 0;
unsigned char highofftab  = 0;
unsigned char lowoff      = 0;
unsigned char highoff     = 0;

unsigned char para2 [0x0F+10]  = {0};
unsigned char lowofftab2   = 0;
unsigned char highofftab2  = 0;
unsigned char lowoff2      = 0;
unsigned char highoff2     = 0;

/* Output file */
FILE *theFile;

/*----------------------------------------------------------------------------------------------------------*/
void outByte(unsigned char ch, bool init = false, int bxinit = 0) {
    static int bx = 0;
    if (init) { 
        bx = bxinit;
    }
    if (!init) {
        printf("%2.2X ", ch);
        bx++;
        if(bx % 16 == 0) {
            printf("\n");
            printf("%8.8X  ", bx);
        }
    } else {
        printf("%8.8X  ", bx);
    }
}


int searchNormal(int &version, int &errCode, string &errText) 
{
    //                         LDA   #00   STA   $ED   STA   $EE   STA   $E9   STA   $EA   JSR
    unsigned char loadCave1[] {0xA9, 0x00, 0x85, 0xED, 0x85, 0xEE, 0x85, 0xE9, 0x85, 0xEA, 0x20};
    
    version = 0;
    errCode = 0;
    errText.clear();

    int pc = 0;
    int startProg = 0x0800;
    int endProg   = 0xD000;

    lowofftab   = 0;
    highofftab  = 0;
    lowoff      = 0;
    highoff     = 0;

    for(pc=startProg; pc<endProg; pc++) {
        unsigned char* rpc = &prog[pc];
        if (memcmp(rpc, loadCave1, sizeof(loadCave1)) == 0) {
            printf("found : %8.8X\n", pc);
            break;
        }
    }
    if (pc < endProg) {
        pc = pc - 15;
        if (prog[pc] == 0x20) 
        {
            unsigned char low         = prog[pc+1];
            unsigned char high        = prog[pc+2];
            printf("copy  : %8.8X\n", high*256+low);

            lowofftab   = 0;
            highofftab  = 0;
            lowoff      = 0;
            highoff     = 0;

            pc = high*256+low;
        
            if(prog[pc+0] == 0xA5 && prog[pc+1] == 0x5D &&
               prog[pc+2] == 0xC5 && prog[pc+3] == 0x99) 
            {
                pc = pc+6;
                for(int cx=0; cx<32; cx++) {
                    if (prog[pc+cx] == 0xA9) {
                        if (highoff == 0) {
                            highoff = prog[pc+cx+1];
                        }
                        if (lowoff == 0) {
                            lowoff  = prog[pc+cx+1];
                            highoff = 0;
                        }
                    }
                    if (prog[pc+cx] == 0x7D) {
                        if (lowofftab == 0 && highofftab == 0) {
                            lowofftab  = prog[pc+cx+1];
                            highofftab = prog[pc+cx+2];
                        }
                    }
                }
                if (lowofftab == 0 && highofftab == 0) {
                    errCode = 4;
                    errText = "-E- lowofftab, highofftab = 0";
                    pc = 0;
                }
            } else {
                errCode = 3;
                errText = "-E- 0x5D,0x99 not found";
                pc = 0;
            }
        } else {
            errCode = 2;
            errText = "-E- JSR copy - not found";
            pc = 0;
        }
    } else {
        errCode = 1;
        errText = "-E- Goon - not found";
        pc = 0;
    }
    return pc;
}


int searchNormal2(int &version, int &errCode, string &errText) 
{
    //                         LDA   #00   STA   $EF   STA   $F0   STA   $EB   STA   $EC   JSR
    unsigned char loadCave1[] {0xA9, 0x00, 0x85, 0xEF, 0x85, 0xF0, 0x85, 0xEB, 0x85, 0xEC, 0x20};
    
    version = 2;
    errCode = 0;
    errText.clear();

    int pc = 0;
    int startProg = 0x0800;
    int endProg   = 0xD000;

    lowofftab   = 0;
    highofftab  = 0;
    lowoff      = 0;
    highoff     = 0;

    for(pc=startProg; pc<endProg; pc++) {
        unsigned char* rpc = &prog[pc];
        if (memcmp(rpc, loadCave1, sizeof(loadCave1)) == 0) {
            printf("found : %8.8X\n", pc);
            break;
        }
    }
    if (pc < endProg) {
        pc = pc - 12;
        if (prog[pc] == 0x20) 
        {
            unsigned char low         = prog[pc+1];
            unsigned char high        = prog[pc+2];
            printf("copy  : %8.8X\n", high*256+low);

            lowofftab   = 0;
            highofftab  = 0;
            lowoff      = 0;
            highoff     = 0;

            pc = high*256+low;
        
            //                 LDA                   0x5D
            //                 ASL                   TAX                   DEX                   DEX
            if(!(prog[pc+0] == 0xA5 && prog[pc+1] == 0x5D &&
                 prog[pc+2] == 0x0A && prog[pc+3] == 0xAA && prog[pc+4] == 0xCA && prog[pc+5] == 0xCA))
            {
                pc = 0xAB17;  // try
            }
            if(prog[pc+0] == 0xA5 && prog[pc+1] == 0x5D &&
               prog[pc+2] == 0x0A && prog[pc+3] == 0xAA && prog[pc+4] == 0xCA && prog[pc+5] == 0xCA) 
            {
                // para
                pc = pc+10;   //     LDA $xxxx,X            // BD02:: start AB17 -> search AB21 -> found AB22 (cx=1) //
                for(int cx=0; cx<32; cx++) {
                    if (prog[pc+cx] == 0xBD) {
                        if (lowofftab == 0) {
                            lowofftab   = prog[pc+cx+1];
                            highofftab  = prog[pc+cx+2];    // BD02:: 89B0
                            break;
                        }
                    }
                }
                if (lowofftab == 0 && highofftab == 0) {
                    errCode = 4;
                    errText = "-E- lowofftab, highofftab = 0";
                    pc = 0;
                }
                // para2
                pc = pc+22;   //     LDA $xxxx,X            // BD02:: search AB37 -> found AB38 (cx=1) //
                for(int cx=0; cx<32; cx++) {
                    if (prog[pc+cx] == 0xBD) {
                        if (lowofftab2 == 0) {
                            lowofftab2   = prog[pc+cx+1];   // BD02:: 89D8
                            highofftab2  = prog[pc+cx+2];
                            break;
                        }
                    }
                }
                if (lowofftab2 == 0 && highofftab2 == 0) {
                    errCode = 4;
                    errText = "-E- lowofftab2, highofftab2 = 0";
                    pc = 0;
                }
            } else {
                errCode = 3;
                errText = "-E- 0x5D not found";
                pc = 0;
            }
        } else {
            errCode = 2;
            errText = "-E- JSR copy - not found";
            pc = 0;
        }
    } else {
        errCode = 1;
        errText = "-E- Goon - not found";
        pc = 0;
    }
    return pc;
}

int main (int argc, char* argv[])
{
    bool doWrite = false;
    bool specialCommand0x0F = false;
    int  caveNoCorrDiff     = 0;
    theFile = stdout;

    string filename;
    string path       = "C:\\temp\\BoulderDashTosec\\_d64\\";
    string fileprefix;

    //--
    // doWrite = true;
    //--
    fileprefix = "Boulder Change (1987)(Liepa, P.)";                                    //--OK--//
    fileprefix = "Boulder Dash - Rocky (1984)(First Star)(es)[h Load 'n' Run]";         //--OK--//  // seems BD01
    fileprefix = "Boulder Dash (1984)(First Star)[cr Nova]";                            //--OK--//  // is BD01   
    fileprefix = "Boulder Dash 01 (1984)(First Star)[cr Fall Guys]";                    //--OK--//  // Boulderdash IV -> but this is not Boulderdash04 !! and not BD01 !!
    fileprefix = "Boulder Dash 02 (1986)(First Star)";                                  //--OK--//  // BD02 format  -- ok
    fileprefix = "Boulder Dash 03 (1986)(First Star)";                                  //--OK--//
    fileprefix = "Boulder Dash 04 (1986)(First Star)";                                  //--OK--//
    //// fileprefix = "Boulder Dash 04 (1986)(First Star)[cr NO]";      // CAV format
    //// fileprefix = "Boulder Dash 04 (1986)(First Star)[cr NO][a]";   // CAV format
    fileprefix = "Boulder Dash 05 (1986)(First Star)[cr 5211]";                         //--OK--//
    fileprefix = "Boulder Dash 06 (1986)(First Star)[cr 5211]";                         //--OK--//
    fileprefix = "Boulder Dash 06 (1986)(First Star)[h Don Pedro][a5]";                 // corrupt!!
    fileprefix = "Boulder Dash 06 (1986)(First Star)[h RCC]";

    //--
    fileprefix = "De Luxe Dash 2 (1990)(Peter Liepa & Marek Roth)[a]--CAVES1";          //--OK--//
    fileprefix = "De Luxe Dash 2 (1990)(Peter Liepa & Marek Roth)[a]--CAVES2";          //--OK--//
    fileprefix = "De Luxe Dash 2 (1990)(Peter Liepa & Marek Roth)[a]--CAVES3";          //--OK--//
    fileprefix = "De Luxe Dash 2 (1990)(Peter Liepa & Marek Roth)[a]--CAVES4";          //--OK--//

    // fileprefix = "De Luxe Dash 3 (1990)(Peter Liepa & Marek Roth)[a]--DREAM3";       //--OK--//
    // specialCommand0x0F = true;

    fileprefix = "Final Boulder 04 (1991)(Liepa, P.)";                                  //--OK--//

    fileprefix = "Profi Boulder 043 (1992)(Peter Liepa & Rockford)";                    //--OK--//  // BD02 format

    fileprefix = "Profi Boulder 087 (1994)(Peter Liepa & Rockford)";
    fileprefix = "Profi Boulder 100 (1995)(Peter Liepa & Rockford)";
goto weiter; 

    //--
weiter:

    if (argc > 1) {
        fileprefix = argv[1];
    }
    if (argc > 2) {
        doWrite = atoi(argv[2]);
    }

    filename = path + fileprefix + ".raw";
    printf("== read file: %s\n", filename.c_str());
    
    /*-- determine mainname --*/
    string prefixname = "";
    size_t last_idx   = filename.rfind('.');
    if (std::string::npos != last_idx) {
        prefixname = filename.substr(0,last_idx);
    } else {
        prefixname = filename;
    }
    string mainname = "";
    last_idx   = prefixname.rfind('\\');
    if (std::string::npos != last_idx) {
        mainname = prefixname.substr(last_idx+1);
    } else {
        mainname = prefixname;
    }
    string mainnameshort;
    string year;
    string author;
    string country;
    string other;
    int    bcnt = 0;
    bool   addch = true;
    for(size_t i=0; i<mainname.size(); i++){
        if (mainname[i] == '(') {
            addch = false;  
            bcnt++;
        }
        if (addch) {
            if (mainname[i] != '#') {   // comment
                mainnameshort += mainname[i];
            }
        } else {
            if (!(mainname[i] == '(' || mainname[i] == ')')) {
                switch (bcnt) {
                    case 1 : year    += mainname[i]; break;
                    case 2 : author  += mainname[i]; break;
                    case 3 : country += mainname[i]; break;
                    default: other   += mainname[i]; break;
                }
            }
        }
        if (mainname[i] == ')') {
            addch = true;  
        }
    }
    /*--*/

    FILE* fp = fopen(filename.c_str(), "rb");
    if (fp!=NULL){
        while(!feof(fp)){
            int rx  = 0;
            rx = fread(prog,1,0xFFFF,fp);
            if (rx == 0 || rx>=0xFFF0){
                break;
            }
        }
        fclose(fp);
    } else {
        printf("-E- File not found: %s\n", filename.c_str());
        return 0;
    }

    int offpara  = 0x20;
    int maxpara  = 0x1FF;
    int offpara2 = 0x00;
    int maxpara2 = 0x05;
    int version  = 0;
    int errCode  = 0;
    string errText = "";

    DecodeCave* decodeCave = NULL;
    decodeCave = new DecodeCaveCommon;
    ((DecodeCaveCommon*)decodeCave)->SetFeature(c_feature_command0x0F, specialCommand0x0F);

    int levMaxEasy = 0;
    int levMaxUser = 0;
    int maxcave    = 20;

    int pc = searchNormal(version, errCode, errText);
    if (pc == 0) {
        version = 2;
        pc = searchNormal2(version, errCode, errText);
        offpara = 0x1A;
        delete decodeCave;
        decodeCave = new DecodeCaveBD02;
        caveNoCorrDiff = -1;
    }
    if (pc > 0) {
        int offset    = highoff*256+lowoff;         // 0x582E
        int offsettab = highofftab*256+lowofftab;   // 0x5804;
        printf("offset : %8.8X\n", offset);
        printf("offtab : %8.8X\n", offsettab);
        printf("version: %d\n", version);
        printf("\n");

        for (int caveNo=1; caveNo<=maxcave; caveNo++) {
        // for (int caveNo=1; caveNo<=1; caveNo++) {
            int caveNoCorr = caveNo + caveNoCorrDiff;  // CHECK
            unsigned char low  = prog[offsettab+caveNoCorr*2];
            unsigned char high = prog[offsettab+caveNoCorr*2+1];
            int off = high*256+low;
            int cavStart = offset + off;
            printf("\n== cave %d -- %c -- off: %8.8X cavStart: %8.8X\n", caveNo, caveNo+'A'-1, off, cavStart);

            // para
            memset(para, 0xFF, sizeof(para));
            outByte(0, true, 0x2400);
            int caveDataIndex = 0;
            for (caveDataIndex = 0; caveDataIndex < offpara; caveDataIndex++) {
                para[caveDataIndex] = prog[cavStart+caveDataIndex];
                outByte(prog[cavStart+caveDataIndex]);
            }
            if (version == 2) {
                int offsettab2 = highofftab2*256+lowofftab2;   
                unsigned char low2  = prog[offsettab2+caveNoCorr*2];
                unsigned char high2 = prog[offsettab2+caveNoCorr*2+1];
                int off2 = high2*256+low2;
                int cavStart2 = off2;
                // para2
                printf("\n");
                printf("offsettab2: %8.8X\n", offsettab2);
                printf("off2      : %8.8X\n", off2);
                memset(para2, 0xFF, sizeof(para2));
                outByte(0, true, 0x24FA);
                int caveDataIndex2 = 0;
                for (caveDataIndex2 = 0; caveDataIndex2 < maxpara2; caveDataIndex2++) {
                    para2[caveDataIndex2] = prog[cavStart2+caveDataIndex2];
                    outByte(prog[cavStart2+caveDataIndex2]);
                }
            }
            // command
            int spara = 0x2400 + offpara;
            // command
            for (caveDataIndex = offpara; caveDataIndex <= maxpara && prog[cavStart+caveDataIndex] != 0xFF; caveDataIndex++) {
                printf("\n%8.8X  ", spara);
                //                 
                UBYTE theCode     = prog[cavStart+caveDataIndex]; 
                UBYTE cmd         = decodeCave->GetCommand(theCode);
                UBYTE cmdLen      = decodeCave->GetCommandLength(cmd);
                UBYTE cmdExtraLen = 0;

                for (int c=0; c < cmdLen; c++) {
                    para[caveDataIndex] = prog[cavStart+caveDataIndex];     // begin with theCode
                    UBYTE byte = para[caveDataIndex];
                    printf("%2.2X ", byte);
                    if (version == 2) {
                        if (theCode == 0x05 && c == 2) {
                            cmdExtraLen = byte;
                        }
                    }
                    if (c < cmdLen-1) {
                        ++caveDataIndex;
                    }
                }
                if (cmdLen >= 10) {
                    printf("    theCode: %2.2X cmd: %2.2X cmdLen: %d", theCode, cmd, cmdLen);
                } else {
                    printf("%*.*s    theCode: %2.2X cmd: %2.2X cmdLen: %d", 30-cmdLen*3, 30-cmdLen*3, " ", theCode, cmd, cmdLen);
                }
                spara = spara + cmdLen;
                if (cmdExtraLen > 0) {
                    printf("\n%8.8X  ", spara);
                    caveDataIndex++;
                    for (int c=0; c < cmdExtraLen; c++) {
                        para[caveDataIndex] = prog[cavStart+caveDataIndex];     // begin with theCode
                        UBYTE byte = para[caveDataIndex];
                        printf("%2.2X ", byte);                
                        if (c < cmdExtraLen-1) {
                            ++caveDataIndex;
                        }
                    }
                    spara = spara + cmdExtraLen;
                }
            }
            printf("\n%8.8X  %2.2X ", spara, 0xFF);
            printf("%*.*s       cave end\n", 30-2*3, 30-2*3, " ");

            //===================================================================================================================== 
            // para with descr output
            unsigned char* aCaveData = &para[0];  // pointer to para
            decodeCave->DoDecodeCave(caveNo+'A'-1, aCaveData, offpara, &prog[0]);  
            printf("\n");

            FILE *fpw = NULL;
            // write file -- create file with header
            if (doWrite) {
                char wfile[256];

                filesystem::path dir1{path+fileprefix};
                filesystem::create_directory(dir1);

                sprintf(wfile, "%s\\%2.2d.WLC", string(path + fileprefix).c_str(), caveNo);
                if (version == 2) {
                    wfile[strlen(wfile)-1] = '2';
                }
                fpw = fopen (wfile, "wb");
                if (fpw != NULL) {
                    unsigned char wbuf[] = {0x00, 0xC1, 'W', 'L', 'C'};
                    if (version == 2) {
                        wbuf[4] = '2';
                    }
                    fwrite(wbuf, 1, sizeof(wbuf), fpw);
                }
            }

            //===================================================================================================================== 
            // cave ausgeben
            printf("\n== cave %d -- %c -- \n", caveNo, caveNo+'A'-1);
            outByte(0, true, 0);
            int cc = 0;
            for (int yy=0;yy<=21;yy++) {
                for (int xx=0;xx<40;xx++) {
                    UBYTE byte = GetRealCaveData(xx, yy+2);
                    outByte(byte);
                    // schreiben                
                    if (doWrite) {
                        if (fpw != NULL) {
                            unsigned char wbuf[] = {byte};
                            fwrite(wbuf, 1, sizeof(wbuf), fpw);
                        }
                    }
                    cc++;
                }
            }
            // write file -- write filler
            if (doWrite) {
                if (fpw != NULL) {
                    unsigned char wbuf[0x10-5] = {0};
                    fwrite(wbuf, 1, sizeof(wbuf), fpw);
                }
            }
            // para
            for (caveDataIndex=0; caveDataIndex<offpara; caveDataIndex++) {
                // schreiben                
                if (doWrite) {
                    if (fpw != NULL) {
                        unsigned char wbuf[] = {para[caveDataIndex]};
                        fwrite(wbuf, 1, sizeof(wbuf), fpw);
                    }
                }
            }
            if (version == 2) {
                /*--*/
                int caveDataIndex2 = 0;
                // para2
                for (caveDataIndex2=0; caveDataIndex2<maxpara2; caveDataIndex2++) {
                    // schreiben                
                    if (doWrite) {
                        if (fpw != NULL) {                                      // bg1,bg2,fg ,
                            unsigned char wbuf[] = {para2[caveDataIndex2]};     // 1A ,1B ,1C ,1D ,1E
                            fwrite(wbuf, 1, sizeof(wbuf), fpw);
                        }
                    }
                }
                // schreiben                
                if (doWrite) {
                    if (fpw != NULL) {
                        unsigned char wbuf[] = {0x00};                          // 1F (filler)
                        fwrite(wbuf, 1, sizeof(wbuf), fpw);
                    }
                }
                /*--*/
            }
            // command
            for (caveDataIndex = offpara; caveDataIndex <= maxpara && para[caveDataIndex] != 0xFF; caveDataIndex++) {
                UBYTE theCode     = para[caveDataIndex];
                UBYTE cmd         = decodeCave->GetCommand(theCode);       // (3 & (theCode >> 6))
                UBYTE cmdLen      = decodeCave->GetCommandLength(cmd);
                UBYTE cmdExtraLen = 0;

                for (int c=0; c < cmdLen; c++) {
                    UBYTE byte = para[caveDataIndex];
                    if (doWrite) {
                        if (fpw != NULL) {
                            unsigned char wbuf[] = {byte};
                            fwrite(wbuf, 1, sizeof(wbuf), fpw);
                        }
                    }
                    if (version == 2) {
                        if (theCode == 0x05 && c == 2) {
                            cmdExtraLen = byte;
                        }
                    }
                    if (c < cmdLen-1) {
                        caveDataIndex++;
                    }
                }
                if (cmdExtraLen > 0) {
                    caveDataIndex++;
                    for (int c=0; c < cmdExtraLen; c++) {
                        UBYTE byte = para[caveDataIndex];
                        if (doWrite) {
                            if (fpw != NULL) {
                                unsigned char wbuf[] = {byte};
                                fwrite(wbuf, 1, sizeof(wbuf), fpw);
                            }
                        }
                        if (c < cmdExtraLen-1) {
                            caveDataIndex++;
                        }
                    }
                }
                cmdLen = 0;
            }
            // schreiben                
            if (doWrite) {
                if (fpw != NULL) {
                    unsigned char wbuf[] = {0xFF};
                    fwrite(wbuf, 1, sizeof(wbuf), fpw);
                }
            }

            // write file -- write filler
            if (doWrite) {
                if (fpw != NULL) {
                    if (version == 2) {
                        /*--*/
                        caveDataIndex += maxpara2+1;  // para2 + filler
                        /*--*/
                    }
                    if (caveDataIndex > offpara) {
                        int fillb = 15 - (caveDataIndex)%16;
                        if (fillb == 0) {
                            fillb = 16;
                        }
                        for(int b=0; b<fillb; b++) {
                            unsigned char wbuf[] = {0xFF};
                            fwrite(wbuf, 1, sizeof(wbuf), fpw);
                        }
                    }
                    fclose(fpw);
                }
            }

            printf("\n\n");
        }

        if (doWrite) {
            string levname = string(path + fileprefix + string("\\") + "levelinfo.conf");
            FILE* fpw = fopen(levname.c_str(), "w");
            string buf;

            cout << "--- Write levelinfo.conf ->" 
                    << " levMax: "  << std::to_string(maxcave) 
                    << " levEasy: " << std::to_string(levMaxEasy) 
                    << " name: " << mainnameshort << endl;

            if (fpw != NULL) {
                buf = "file_identifier:                ROCKSNDIAMONDS_LEVELINFO_FILE_VERSION_1.2\n";
                fputs(buf.c_str(), fpw);
                buf = "\n";
                fputs(buf.c_str(), fpw);
                buf = "name:                           " + mainnameshort + "\n"; 
                fputs(buf.c_str(), fpw);
                buf = "author:                         " + author + "\n"; 
                fputs(buf.c_str(), fpw);
                buf = "\n";
                fputs(buf.c_str(), fpw);
                buf = "levels:                         " + std::to_string(maxcave) + "\n";
                fputs(buf.c_str(), fpw);
                buf = "first_level:                    " + string("1") + "\n";
                fputs(buf.c_str(), fpw);
                buf = "readonly:                       " + string("true") + "\n"; 
                fputs(buf.c_str(), fpw);
                buf = "\n";
                fputs(buf.c_str(), fpw);

                buf = "#mainname:                      " + mainname + "\n"; 
                fputs(buf.c_str(), fpw);
                buf = "#year:                          " + year + "\n"; 
                fputs(buf.c_str(), fpw);
                buf = "#country:                       " + country + "\n"; 
                fputs(buf.c_str(), fpw);
                buf = "#other:                         " + other + "\n"; 
                fputs(buf.c_str(), fpw);
                buf = "#levMaxSelect:                  " + std::to_string(levMaxEasy) + "\n"; 
                fputs(buf.c_str(), fpw);

                buf = "\n";
                fputs(buf.c_str(), fpw);
                fclose(fpw);
            }
        }

    } else {
        printf("WARNING - normal code not found! %.3d %s\n", errCode, errText.c_str());
    }

    if (doWrite) {
        // write/append CSV
        if (maxcave >= 0)  // immer schreiben
        {
            FILE* fpw = fopen("C:\\temp\\BoulderDashTosecReaderWLC.csv", "a+");
            string buf;
            buf = mainname + ";" + 
                  mainnameshort + ";" + 
                  ";" +   // remark
                  ";" +   // check
                  std::to_string(maxcave) + ";" + 
                  std::to_string(levMaxUser) + ";" + 
                  std::to_string(levMaxEasy) + ";" +
                  author  + ";" + 
                  year    + ";" +
                  country + ";" +
                  other   + ";" +
                  "\n";

            if (fpw != NULL) {
                fputs(buf.c_str(), fpw);
                fclose(fpw);
            }
        }
    }


    printf("\n");
    printf("\n");
}
