// ============================================================================
// decodeCave.h
// ----------------------------------------------------------------------------
// (c) 2022-2024 by HerzAusGold
//               Hubert Haberkorn
//               herz-aus-gold-42@gmx.de
//               http://www.arthurdent.de/
// ----------------------------------------------------------------------------
// The above copyright notice and this permission notice 
// shall be included in all copies or substantial portions 
// of the Software.
// ----------------------------------------------------------------------------
// read "raw" file dump and extract the data with parameters.
// create .WLC file - WithLevelCommon
// ----------------------------------------------------------------------------

#ifndef DECODECAVE_H
#define DECODECAVE_H

/* **************************************** */
/* Types */
typedef unsigned char UBYTE;
typedef short objectType;

/* Global storage for the decoded cave data */
#define MAXX  40
#define MAXY  24

/* **************************************** */
#define INCLUDE_HUMAN_READABLE  1

UBYTE * GetRealCaveDataPointer(int x, int y);

char GetCreatorCode(UBYTE ch);
const char * GetDirections(UBYTE dir);
const char * GetObjects(UBYTE obj);
const char * GetColors(UBYTE col);
UBYTE GetCaveData(int x, int y);
UBYTE GetRealCaveData(int x, int y);
void StoreObject(int x, int y, objectType anObject);
void StoreObjectDirect(int addr, objectType anObject);

void DrawLine(objectType anObject, int x, int y, int aLength, int aDirection);
void DrawFilledRectSimple(objectType anObject, int x, int y, int aWidth, int aHeight);
void DrawFilledRect(objectType anObject, int x, int y, int aWidth, int aHeight, objectType aFillObject);
void DrawRect(objectType anObject, int x, int y, int aWidth, int aHeight);
void NextRandom(int *RandSeed1, int *RandSeed2);

class DecodeCave {
    public:
        virtual UBYTE GetCommand(UBYTE cmd) = 0;
        virtual UBYTE GetCommandLength(UBYTE cmd) = 0;
        virtual void DoDecodeCave(unsigned char caveNo, UBYTE *aCaveData, int offpara, UBYTE *prog) = 0;

};

#endif
