// ============================================================================
// decodeCaveBD02.cpp
// ----------------------------------------------------------------------------
// (c) 2022-2024 by HerzAusGold
//               Hubert Haberkorn
//               herz-aus-gold-42@gmx.de
//               http://www.arthurdent.de/
// ----------------------------------------------------------------------------
// The above copyright notice and this permission notice 
// shall be included in all copies or substantial portions 
// of the Software.
// ----------------------------------------------------------------------------
// read "raw" file dump and extract the data with parameters.
// create .WLC file - WithLevelCommon
// ----------------------------------------------------------------------------

#include "stdio.h"
#include <vector>
#include "decodeCave.h"
#include "decodeCaveBD02.h"


extern FILE* theFile;


/* **************************************** */
UBYTE DecodeCaveBD02::GetCommand(UBYTE cmd)
{
    return cmd;
}

/* **************************************** */
UBYTE DecodeCaveBD02::GetCommandLength(UBYTE cmd)
{
    UBYTE cmdLen = 0;
    switch (cmd) {
        case 0x03: cmdLen = 4; break;  // $AD8D // Plot
        case 0x00: cmdLen = 6; break;  // $AD95 // Line
        case 0x02: cmdLen = 7; break;  // $AD9D // FilledRect
        case 0x01: cmdLen = 6; break;  // $ADA5 // Rect
        case 0x04: cmdLen = 8; break;  // $ADAD // Dotted
        case 0x06: cmdLen = 4; break;  // $ADB5 // Replace with offset // BD02 in Level G // 
        case 0x05: cmdLen = 5; break;  // $ADBD // BitSet
        case 0x07: cmdLen = 2; break;  // $ADC2 // Move $DA -> $99  // 0x07 0xE0 -- 0xE0 -> $99 // BD02 in Level E, M 
        case 0x08: cmdLen = 0; break;  // $ADCA // ??? UNKNOWN ???
        case 0x09: cmdLen = 5; break;  // $5B26 // NIBBLE -- posy, posx for player
        case 0xFE: cmdLen = 1; break;  // next $#E0 bytes are commands
        default:cmdLen = 4; break;     // $ABBE // 
    }
    return cmdLen;
}


static void PrintCave(unsigned char caveNo) {
    int y,x;
    fprintf(theFile,"\n\nMap %c #%d\n",caveNo, caveNo-'A'+1);
    fprintf(theFile,"--- 0         1         2         3         4\n");
    fprintf(theFile,"--- 01234567890123456789012345678901234567890\n");
    for (y = 2; y <= 23; y++) {
        fprintf(theFile, "%03d ", y);
        for(x = 0; x <= 39; x++) {
            fprintf(theFile, "%c", GetCaveData(x, y));  // get them
        }
        if (y == 23) {
            fprintf(theFile, "");
        }
        fprintf(theFile, "\n");
    }
    y=0;
    x=0;
}


/* **************************************** */
void DecodeCaveBD02::DoDecodeCave(unsigned char caveNo, UBYTE *aCaveData, int offpara, UBYTE *prog)
{
    int RandSeed1, RandSeed2;
    int theWidth, theHeight, theFill, theLength, theDirection;
    int theXCount, theYCount, theXSpace, theYSpace;
    int theReplace, theOffset;
    int theStartAddr;
    int x, y;
    int caveDataIndex;
    UBYTE theObject;
    UBYTE theByte;
    int theCode;
    //------------------------------
    int diamondWorth    = 0x01;
    int diamondExWorth  = 0x02;
    int timeNeeded      = 0x03;   // 0x03,0x04,0x05,0x06,0x07 
    int diamondNeeded   = 0x08;   // 0x08,0x09,0x0A,0x0B,0x0C
    //------------------------------
    int randSeedStart   = 0x0D;   // 0x0D,0x0E,0x0F,0x10,0x11
    int randPerStart    = 0x12;   // 0x12,0x13,0x14,0x15
    int randObjStart    = 0x16;   // 0x16,0x17,0x18,0x19
    //------------------------------
    // offpara = 0x19; 
    //------------------------------

    RandSeed1 = 0;
    RandSeed2 = aCaveData[randSeedStart];   // aCaveData[4];    // random seed

/* Output human-readable version of the first 0x20 (offpara) bytes of cave data. */
    #if INCLUDE_HUMAN_READABLE
        fprintf(theFile,"\n\n=== Cave %d (Cave %c)\n\n",caveNo-'A'+1,caveNo);

        fprintf(theFile,"%.2X                  Magic wall/amoeba slow growth for: %d seconds\n",aCaveData[0],aCaveData[0]);
        fprintf(theFile,"%.2X                  Diamonds worth: %d points\n",aCaveData[diamondWorth],aCaveData[diamondWorth]);
        fprintf(theFile,"%.2X                  Extra diamonds worth: %d points\n",aCaveData[diamondExWorth],aCaveData[diamondExWorth]);

        // fprintf(theFile,"%.2X                  Magic wall/amoeba slow growth for: %d seconds\n",aCaveData[3],aCaveData[3]);


        fprintf(theFile,"%.2X %.2X %.2X %.2X %.2X      Cave time: %d, %d, %d, %d, %d seconds\n",
                        aCaveData[timeNeeded+0],aCaveData[timeNeeded+1],aCaveData[timeNeeded+2],aCaveData[timeNeeded+3],aCaveData[timeNeeded+4],
                        aCaveData[timeNeeded+0],aCaveData[timeNeeded+1],aCaveData[timeNeeded+2],aCaveData[timeNeeded+3],aCaveData[timeNeeded+4]);
        fprintf(theFile,"%.2X %.2X %.2X %.2X %.2X      Diamonds needed: %d, %d, %d, %d, %d (for difficulty levels 1-5)\n",
                        aCaveData[diamondNeeded+0],aCaveData[diamondNeeded+1],aCaveData[diamondNeeded+2],aCaveData[diamondNeeded+3],aCaveData[diamondNeeded+4],
                        aCaveData[diamondNeeded+0],aCaveData[diamondNeeded+1],aCaveData[diamondNeeded+2],aCaveData[diamondNeeded+3],aCaveData[diamondNeeded+4]);

        // random seed 
        fprintf(theFile,"%.2X %.2X %.2X %.2X %.2X      Randomiser seed values for difficulty levels 1-5\n",
                        aCaveData[randSeedStart+0],aCaveData[randSeedStart+1],aCaveData[randSeedStart+2],aCaveData[randSeedStart+3],aCaveData[randSeedStart+4]);

        // random section
        if (aCaveData[randPerStart+0]) {
            fprintf(theFile,"%.2X %.2X %.2X %.2X             %-7s: %3d/256 = %2d%%\n",
                    aCaveData[randPerStart+0],aCaveData[randPerStart+1],aCaveData[randPerStart+2],aCaveData[randPerStart+3], 
                    GetObjects(aCaveData[randObjStart+0]),aCaveData[randPerStart+0],(int)aCaveData[randPerStart+0]*100/256);
        } else {
            fprintf(theFile,"%.2X %.2X %.2X %.2X             first  code unused (0%%)\n",
                    aCaveData[randPerStart+0],aCaveData[randPerStart+1],aCaveData[randPerStart+2],aCaveData[randPerStart+3]);
        }
        if (aCaveData[randPerStart+1]) {
            fprintf(theFile,"                        %-7s: %3d/256 = %2d%%\n", 
                    GetObjects(aCaveData[randObjStart+1]),aCaveData[randPerStart+1],(int)aCaveData[randPerStart+1]*100/256);
        } else {
            fprintf(theFile,"                        second code unused (0%%)\n");
        }
        if (aCaveData[randPerStart+2]) {
            fprintf(theFile,"                        %-7s: %3d/256 = %2d%%\n", 
                    GetObjects(aCaveData[randObjStart+2]),aCaveData[randPerStart+2],(int)aCaveData[randPerStart+2]*100/256);
        } else {
            fprintf(theFile,"                        third  code unused (0%%)\n");
        }
        if (aCaveData[randPerStart+3]) {
            fprintf(theFile,"                        %-7s: %3d/256 = %2d%%\n", 
                    GetObjects(aCaveData[randObjStart+3]),aCaveData[randPerStart+3],(int)aCaveData[randPerStart+3]*100/256);
        } else {
            fprintf(theFile,"                        fourth code unused (0%%)\n");
        }
        fprintf(theFile,"%.2X %.2X %.2X %.2X         Random objects:\n",
                        aCaveData[randObjStart+0],aCaveData[randObjStart+1],aCaveData[randObjStart+2],aCaveData[randObjStart+3]);

        fprintf(theFile,"Cave layout:\n");
    #endif

/* Clear out the cave data to a null value */
    for(x = 0; x < 40; x++) {
        for (y = 0; y <= 23; y++) {
            StoreObject(x, y, 0x07);
        }
    }

/* Decode the random cave data */
    for(y = 3; y <= 23; y++) {
        for(x = 0; x <= 39; x++) {
            theObject = 1;  /* Dirt */
            NextRandom(&RandSeed1, &RandSeed2);
            for (caveDataIndex = 0; caveDataIndex <= 3; caveDataIndex++) {
                if (RandSeed1 < aCaveData[randPerStart + caveDataIndex]) {
                    theObject = aCaveData[randObjStart + caveDataIndex];
                }
            }
            StoreObject(x, y, theObject);
        }     
    }  

/* SteelBounds */
    DrawRect(0x07, 0, 2, 40, 22);  // before commands!! // x,y,width,height

/* Decode the explicit cave data (commands) */
/* start from "offpara" (usually 0x20)      */
    for (caveDataIndex = offpara; aCaveData[caveDataIndex] != 0xFF; caveDataIndex++) 
    {
        // PrintCave(caveNo);
        //----------------------------------------------------------------
        theCode   = aCaveData[caveDataIndex];
        // fprintf(theFile, "\n%2.2X ", theCode);

        UBYTE cmd    = GetCommand(theCode);
        UBYTE cmdLen = GetCommandLength(cmd);

        std::vector<UBYTE> valArrCmd;
        for(int cx=0; cx<cmdLen; cx++) {
            UBYTE val = aCaveData[caveDataIndex];   // begin with theCode
            valArrCmd.push_back(val);
            fprintf(theFile, "%2.2X ", val);
            if (cx < cmdLen-1) {
                caveDataIndex++;
            }
        }
        if (cmdLen != valArrCmd.size()) {
            cmd = 0xFF;  // break gere
        }

        switch(cmd) {   
            case 0x00   : /* LINE */
                theByte      = valArrCmd[0];
                theObject    = valArrCmd[1];
                y            = valArrCmd[2];
                x            = valArrCmd[3];
                theLength    = valArrCmd[5];
                // "up","up/right","right","down/right","down","down/left","left","up/left"
                switch(valArrCmd[4]) {
                    case 0x00 : theDirection = 0;  break;  // up
                    case 0x02 : theDirection = 1;  break;  // up/right
                    case 0x04 : theDirection = 2;  break;  // right
                    case 0x06 : theDirection = 3;  break;  // down/right
                    case 0x08 : theDirection = 4;  break;  // down
                    case 0x0A : theDirection = 5;  break;  // down/left
                    case 0x0C : theDirection = 6;  break;  // left
                    case 0x0E : theDirection = 7;  break;  // up/left
                    default   : theDirection = 0;  break;  //-- set breakpoint here!!
                }
                #if INCLUDE_HUMAN_READABLE
                    fprintf(theFile,"\n--> %.2X %.2X %.2X %.2X %.2X %.2X         Line of %s from (%2d,%2d); direction = %s; length = %d\n",
                                    theCode,theObject,y,x,valArrCmd[3],theLength,
                                    GetObjects(theObject),y+2,x,GetDirections(theDirection),theLength);
                #endif
                DrawLine(theObject, x, y+2, theLength, theDirection);
                break;

            case 0x03   : /* PLOT */
                theByte      = valArrCmd[0];
                theObject    = valArrCmd[1];
                y            = valArrCmd[2];
                x            = valArrCmd[3];
                #if INCLUDE_HUMAN_READABLE
                    fprintf(theFile,"\n--> %.2X %.2X %.2X %.2X               StoreChar %s at (%d,%d)\n",
                                    theCode,theObject,y,x,
                                    GetObjects(theObject),y+2,x);
                #endif
                StoreObject(x, y+2, theObject);
                break;

            case 0x02   : /* FILLED RECT */
                theByte      = valArrCmd[0];
                theObject    = valArrCmd[1];
                y            = valArrCmd[2];
                x            = valArrCmd[3];
                theHeight    = valArrCmd[4];
                theWidth     = valArrCmd[5];
                theFill      = valArrCmd[6];
                #if INCLUDE_HUMAN_READABLE
                    fprintf(theFile,"\n--> %.2X %.2X %.2X %.2X %.2X %.2X %.2X      FilledRect of %s from (%d,%d); width = %d; height = %d; fill = %s\n",
                            theCode,theObject,y,x,theHeight,theWidth,theFill,
                            GetObjects(theObject),y+2,x,theWidth,theHeight,GetObjects(theFill));
                #endif
                DrawFilledRect(theObject, x, y+2, theWidth, theHeight, theFill);
                break;
            
            case 0x04   : /* DOTTED */
                theByte      = valArrCmd[0];
                theObject    = valArrCmd[1];
                y            = valArrCmd[2];
                x            = valArrCmd[3];
                theYCount    = valArrCmd[4];
                theXCount    = valArrCmd[5];
                theYSpace    = valArrCmd[6];
                theXSpace    = valArrCmd[7];
                #if INCLUDE_HUMAN_READABLE
                    fprintf(theFile,"\n--> %.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X   Dotted of %s from (%2d,%2d); Ycount = %d Xcount = %d; YSpace = %d; XSpace = %d\n",
                            theCode,theObject,y,x,theYCount, theXCount, theYSpace, theXSpace,
                            GetObjects(theObject),y+2,x,theYCount,theXCount,theYSpace,theXSpace);
                #endif
                /*--
                for(int iy=0;iy<theYCount;iy++) {
                    x = valArrCmd[3];
                    for(int ix=0;ix<theXCount;ix++) {
                        StoreObject(x, y+2, theObject);
                        x = x + theXSpace;
                    }
                    y = y + theYSpace;
                }
                --*/
                for(int iy=0;iy<theYCount;iy++) {
                    int xx = x;
                    for(int ix=0;ix<theXCount;ix++) {
                        StoreObject(xx, y+2, theObject);
                        xx = xx + theXSpace;
                        if (xx >= 40) {
                            xx = xx - 40;
                            y  = y + 1;
                        }
                    }
                    y = y + theYSpace;
                }
                break;

            case 0x06   : /* Replace with Offset */
                theByte      = valArrCmd[0];
                theObject    = valArrCmd[1];
                theReplace   = valArrCmd[2];
                theOffset    = valArrCmd[3];
                #if INCLUDE_HUMAN_READABLE
                {
                    int dx = theOffset % 40;
                    int dy = theOffset / 40; 
                    fprintf(theFile,"\n--> %.2X %.2X %.2X %.2X               Replace of %s with offset: %d (dx:%d dy:%d) to %s\n",
                            theCode,theObject,theReplace,theOffset,
                            GetObjects(theObject), theOffset, dx, dy, GetObjects(theReplace));
                }
                #endif
                /*--
                for(y=2; y<MAXY-1; y++) {
                    for(x=0; x<MAXX; x++) {
                        if (GetRealCaveData(x,y) == theObject) {
                            int newx = x + theOffset;
                            int newy = y;
                            StoreObject(newx, newy, theReplace);
                        }
                    }
                }
                --*/
                for(y=2; y<MAXY-1; y++) {
                    for(x=0; x<MAXX; x++) {
                        if (GetRealCaveData(x,y) == theObject) {
                            int dx   = theOffset % 40;
                            int dy   = theOffset / 40;
                            int newx = x + dx;
                            int newy = y + dy;
                            StoreObject(newx, newy, theReplace);
                        }
                    }
                }
                break;

            case 0x01   : /* Rect */
                theByte      = valArrCmd[0];
                theObject    = valArrCmd[1];
                y            = valArrCmd[2];
                x            = valArrCmd[3];
                theHeight    = valArrCmd[4];
                theWidth     = valArrCmd[5];
                #if INCLUDE_HUMAN_READABLE
                    fprintf(theFile,"\n--> %.2X %.2X %.2X %.2X %.2X %.2X         Rect of %s from (%d,%d) height %d width %d\n",
                            theCode,theObject,y,x,theHeight,theWidth,
                            GetObjects(theObject), y+2, x, theHeight, theWidth);
                #endif
                DrawRect(theObject, x, y+2, theWidth, theHeight);
                break;
                
            case 0x07   : /* INTERNAL COMMAND -- Move to $99 */
                theByte      = valArrCmd[0];
                theObject    = valArrCmd[1];
                #if INCLUDE_HUMAN_READABLE
                    fprintf(theFile,"\n--> %.2X %.2X                     (INTERNAL) Move %.2X to $99\n",
                            theCode,theObject,theObject);
                #endif
                break; 

            case 0x05   : /* BitSet */
                {
                    theByte            = valArrCmd[0];
                    theObject          = valArrCmd[1];
                    UBYTE cmdExtraLen  = valArrCmd[2];
                    UBYTE startLow     = valArrCmd[3];
                    UBYTE startHigh    = valArrCmd[4];
                    theStartAddr = startHigh*256+startLow;
                    theStartAddr = theStartAddr - 0x0828 + MAXX;
                    //
                    #if INCLUDE_HUMAN_READABLE
                        fprintf(theFile,"\n--> %.2X %.2X %.2X %.2X %.2X            BitSet for object: %s len: %d startAddr: $%4.4X relativ: $%4.4X\n",
                                theCode,theObject,cmdExtraLen,startLow, startHigh,
                                GetObjects(theObject),cmdExtraLen,startHigh*256+startLow,theStartAddr);
                    #endif
                    //
                    for(int lx=0; lx<cmdExtraLen; lx++) {
                        theByte = aCaveData[++caveDataIndex];
                        fprintf(theFile, "%.2X ", theByte);
                        for(int bx=0; bx<8; bx++) {
                            if (theByte%2 == 1) {
                                StoreObjectDirect(theStartAddr, theObject);
                            }
                            theByte = theByte >> 1;  // LSR
                            theStartAddr++;
                        }
                    }
                    fprintf(theFile,"\n-->                           BitSet end\n");
                }
                break;

            case 0x09   : /* NIBBLE */
                {
                    theByte            = valArrCmd[0];
                    UBYTE startLow     = valArrCmd[1];
                    UBYTE startHigh    = valArrCmd[2];
                    theStartAddr = startHigh*256+startLow;
                    // theStartAddr = theStartAddr - 0x0828 + MAXX;
                    y                  = valArrCmd[3];
                    x                  = valArrCmd[4];
                    int cmdExtraLen    = 400;   // fix 20*40/2
                    #if INCLUDE_HUMAN_READABLE
                        fprintf(theFile,"\n--> %.2X %.2X %.2X %.2X %.2X            Nibble start at: $%4.4X (%d bytes) Player at: (%d,%d)\n",
                                theCode,startLow,startHigh,y,x,
                                startHigh*256+startLow,cmdExtraLen,y,x);
                    #endif

                    UBYTE replObject[16] = {0x10, 0x14, 0x03, 0x02, 0x07, 0x2A, 0x3C, 0x01, 0x08, 0x30, 0x3A, 0x3E, 0x3D, 0x04, 0x25, 0x00};
                    // repl         //        00    01    02    03    04    05    06    07    08    09    0A    0B    0C    0D    0E    0F

                    if (prog != NULL) {
                        for(int lx=0; lx<cmdExtraLen; lx++) {
                            theByte = prog[theStartAddr+lx];
                            UBYTE theHigh = theByte & 0xF0;
                            UBYTE theLow  = theByte & 0x0F;
                            theHigh = theHigh >> 4;
                            theObject = replObject[theHigh];
                            int theAddr = MAXX + MAXX*2 + lx*2;
                            StoreObjectDirect(theAddr+0, theObject);
                            theObject = replObject[theLow];
                            StoreObjectDirect(theAddr+1, theObject);
                        }
                    }
                }
                break;

            case 0xFE   : /* #$E0 bytes more commands */
                {
                    theByte            = valArrCmd[0];
                    #if INCLUDE_HUMAN_READABLE
                        fprintf(theFile,"\n--> %.2X                         (INTERNAL) Next #$E0 Bytes are Commands\n",
                                theCode);
                    #endif
                }
                break;

            default     : 
                valArrCmd.clear();     //-- set breakoint here!!
                break;
        }
    }

/* SteelBounds */
//    DrawRect(0x07, 0, 2, 40, 22);  // before commands!!

/* Print out the resulting aCaveData */
    #if INCLUDE_HUMAN_READABLE
        fprintf(theFile,"FF                            End of cave data.");
    #endif
    PrintCave(caveNo);
}


