// ============================================================================
// decodeCaveCommon.cpp
// ----------------------------------------------------------------------------
// (c) 2022-2024 by HerzAusGold
//               Hubert Haberkorn
//               herz-aus-gold-42@gmx.de
//               http://www.arthurdent.de/
// ----------------------------------------------------------------------------
// The above copyright notice and this permission notice 
// shall be included in all copies or substantial portions 
// of the Software.
// ----------------------------------------------------------------------------
// read "raw" file dump and extract the data with parameters.
// create .WLC file - WithLevelCommon
// ----------------------------------------------------------------------------

#include "stdio.h"
#include "decodeCave.h"
#include "decodeCaveCommon.h"

extern FILE* theFile;

#pragma region "BD01 first two levels"

/*--------------------------------------------------*/
/*------- BD01 -------------------------------------*/
/*--- see: decodedCaveData.txt               -------*/
/*--- see: original_decodecaves_c.txt        -------*/
/*--- see: original_cavedata_h.txt           -------*/
/*--------------------------------------------------*/
/*--------------------------------------------------*/

/*--- parameter format --> BD01 cave A
* 0000: 01 -> Cave 01
* 0001: 14 -> Magic wall/amoeba slow growth for: 20 seconds
* 0002: 0A -> Diamonds worth: 10 points
* 0003: 0F -> Extra Diamonds worth: 15 points
* 0004: 0A 0B 0C 0D 0E -> Randomiser seed values for difficulty levels 1-5
* 0009: 0C 0C 0C 0C 0C -> Diamonds needed for difficulty levels 1-5: 12,12,12,12,12
* 000E: 96 6E 46 28 1E -> Cave time: 150, 110, 70, 40, 30 seconds
* 0013: 08 -> Background Color 1: Orange
* 0014: 0B -> Background Color 2: Gray1
* 0015: 09 -> Foreround Color: Brown
* 0016: D4 20 -> Unused
* 0018: 00 10 14 00  -> Random objects: 00 10 14 00 
* 001C: 3C 32 09 00  ->    zSpace: 60/256 = 23%   3C = 60  object 00 -> space
*                    ->    zBouS : 50/256 = 19%   32 = 50  object 10 -> rock
*                    ->    zDiaS :  9/256 =  3%   09 =  9  object 14 -> diamonds
*                    ->    fourth code unused (0%)         object 00 -> space -> with 0% -> unused
* 0020:
Cave layout:
42 01 09 1E 02      Line of zBrick from ( 1, 9); length = 30; direction = right
42 09 10 1E 02      Line of zBrick from ( 9,16); length = 30; direction = right
25 03 04            StoreChar zPRFd1 at ( 3, 4)
04 26 12            StoreChar zPreOut at (38,18)
FF                  End of cave data.

Map #1
002 WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
003 W...... ..d.r .....r.r....... ....r....W
004 W.rXr...... .........rd..r.... ..... ..W
005 W.......... ..r.....r.r..r........r....W
006 Wr.rr.........r......r..r....r...r.....W
007 Wr. r......... r..r........r......r.rr.W
008 W... ..r........r.....r. r........r.rr.W
009 Wwwwwwwwwwwwwwwwwwwwwwwwwwwwwww...r..r.W
010 W. ...r..d. ..r.r..........d.rd...... .W
011 W..d.....r..... ........rr r..r....r...W
012 W...r..r.r..............r .r..r........W
013 W.r.....r........rrr.......r.. .d....r.W
014 W.d.. ..r.  .....r.rd..d....r...r..d. .W
015 W. r..............r r..r........d.....rW
016 W........wwwwwwwwwwwwwwwwwwwwwwwwwwwwwwW
017 W r.........r...d....r.....r...r.......W
018 W r......... r..r........r......r.rr..PW
019 W. ..r........r.....r.  ....d...r.rr...W
020 W....rd..r........r......r.rd......r...W
021 W... ..r. ..r.rr.........r.rd...... ..rW
022 W.d.... ..... ......... .r..r....r...r.W
023 WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW

------------------------------------------------------------------------------------------------------
02                  Cave 02
14                  Magic wall/amoeba slow growth for: 20 seconds
14                  Diamonds worth: 20 points
32                  Extra diamonds worth: 50 points
03 00 01 57 58      Randomiser seed values for difficulty levels 1-5
0A 0C 09 0D 0A      Diamonds needed: 10, 12, 9, 13, 10 (for difficulty levels 1-5)
96 6E 46 46 46      Cave time: 150, 110, 70, 70, 70 seconds
0A                  Background color 1: Light red
04                  Background color 2: Purple
09                  Foreground color: Brown
00 00               Unused
00 10 14 08         Random objects: 00 10 14 08 
3C 32 09 02             zSpace :  60/256 = 23%  3C = 60   object: 00 -> space
                        zBouS  :  50/256 = 19%  32 = 50   object: 10 -> rock
                        zDiaS  :   9/256 =  3%  09 =  9   object: 14 -> diamonds
                        zFFly1 :   2/256 =  0%  02 =  2   object: 08 -> firefly1 -> firefly (links)
Cave layout:
42 01 08 26 02      Line of zBrick from ( 1, 8); length = 38; direction = right
42 01 0F 26 02      Line of zBrick from ( 1,15); length = 38; direction = right
42 08 03 14 04      Line of zBrick from ( 8, 3); length = 20; direction = down
42 10 03 14 04      Line of zBrick from (16, 3); length = 20; direction = down
42 18 03 14 04      Line of zBrick from (24, 3); length = 20; direction = down
42 20 03 14 04      Line of zBrick from (32, 3); length = 20; direction = down
40 01 05 26 02      Line of zSpace from ( 1, 5); length = 38; direction = right
40 01 0B 26 02      Line of zSpace from ( 1,11); length = 38; direction = right
40 01 12 26 02      Line of zSpace from ( 1,18); length = 38; direction = right
40 14 03 14 04      Line of zSpace from (20, 3); length = 20; direction = down
25 12 15            StoreChar zPRFd1 at (18,21)
04 12 16            StoreChar zPreOut at (18,22)
FF                  End of cave data.

Map #2
002 WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
003 W.r..r..w.r...d.w... .r.wr......w..rr..W
004 W.......w......rwrr. ...w ..d...w....r.W
005 W                                      W
006 Wd......w.r....rw.r. .. w..r..d.w..r.r.W
007 W.......w.r....rw.r. r..w.....r.w... ..W
008 Wwwwwwwwwwwwwwwwwwww wwwwwwwwwwwwwwwwwwW
009 W....rr.w..r....w... ..rw....r..w.....rW
010 W.......w.. ....w... ...w....r. w.....rW
011 W                                      W
012 Wr..r...w....r..w..r ...w......dwr.....W
013 Wr....r.w..r..r.w... . rw.......wr...r.W
014 W.r.....w...r...w... . rw.......w r..r.W
015 Wwwwwwwwwwwwwwwwwwww wwwwwwwwwwwwwwwwwwW
016 Wr.  q..w....r.rw... ...w.rd..r.w......W
017 W.....r.wr......w..d ...w ..r...w.r.rr.W
018 W                                      W
019 Wd.. .r.wr....r.w.r. ..rw.r.r...w......W
020 W.....r.wr..d...w... r..w..r....w...rr W
021 W.d... rw..r....w.Xd r..w. .....w...rr W
022 W.r.... w.. ..r.w.P. ...w....r.rw.... .W
023 WWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW

---*/

#pragma endregion


/* **************************************** */
UBYTE DecodeCaveCommon::GetCommand(UBYTE cmd)
{
    return (3 & (cmd >> 6));
}

/* **************************************** */
UBYTE DecodeCaveCommon::GetCommandLength(UBYTE cmd)
{
    UBYTE cmdLen = 0;
    switch (cmd) {
        case 0: cmdLen = 3; break;  // PLOT
        case 1: cmdLen = 5; break;  // LINE
        case 2: cmdLen = 6; break;  // FILLED RECT
        case 3: cmdLen = 5; break;  // RECT
        default:cmdLen = 0; break;
    }
    return cmdLen;
}

/* **************************************** */
void DecodeCaveCommon::DoDecodeCave(unsigned char caveNo, UBYTE *aCaveData, int offpara, UBYTE *prog)
{
    int RandSeed1, RandSeed2;
    int theWidth, theHeight, theFill, theLength, theDirection;
    int theXCount, theYCount, theXSpace, theYSpace;
    int x, y;
    int caveDataIndex;
    UBYTE theObject;
    int theCode;

    RandSeed1 = 0;
    RandSeed2 = aCaveData[4];

/* Output human-readable version of the first 0x20 (offpara) bytes of cave data. */
    #if INCLUDE_HUMAN_READABLE
        if (aCaveData[0] <= 16) {
            fprintf(theFile,"\n\n=== Cave %d (Cave %c)\n\n",aCaveData[0],aCaveData[0]+'A'-1);
                            // "", "");   // caveNames[aCaveData[0]-1],caveDescriptions[aCaveData[0]-1]);
        } else {
            fprintf(theFile,"\n\n=== Cave %d (Intermission %d)\n",aCaveData[0],aCaveData[0]-16);
        }
        fprintf(theFile,"%.2X                  Cave %.2d\n",aCaveData[0],aCaveData[0]);

        fprintf(theFile,"%.2X                  Magic wall/amoeba slow growth for: %d seconds\n",aCaveData[1],aCaveData[1]);
        fprintf(theFile,"%.2X                  Diamonds worth: %d points\n",aCaveData[2],aCaveData[2]);
        fprintf(theFile,"%.2X                  Extra diamonds worth: %d points\n",aCaveData[3],aCaveData[3]);
        fprintf(theFile,"%.2X %.2X %.2X %.2X %.2X      Randomiser seed values for difficulty levels 1-5\n",
                        aCaveData[4],aCaveData[5],aCaveData[6],aCaveData[7],aCaveData[8]);
        fprintf(theFile,"%.2X %.2X %.2X %.2X %.2X      Diamonds needed: %d, %d, %d, %d, %d (for difficulty levels 1-5)\n",
                        aCaveData[9],aCaveData[10],aCaveData[11],aCaveData[12],aCaveData[13],
                        aCaveData[9],aCaveData[10],aCaveData[11],aCaveData[12],aCaveData[13]);
        fprintf(theFile,"%.2X %.2X %.2X %.2X %.2X      Cave time: %d, %d, %d, %d, %d seconds\n",
                        aCaveData[14],aCaveData[15],aCaveData[16],aCaveData[17],aCaveData[18],
                        aCaveData[14],aCaveData[15],aCaveData[16],aCaveData[17],aCaveData[18]);
        fprintf(theFile,"%.2X                  Background color 1: %s\n",aCaveData[19], GetColors(aCaveData[19]));
        fprintf(theFile,"%.2X                  Background color 2: %s\n",aCaveData[20], GetColors(aCaveData[20]));
        fprintf(theFile,"%.2X                  Foreground color  : %s\n",aCaveData[21], GetColors(aCaveData[21]));
        fprintf(theFile,"%.2X %.2X               Unused\n",aCaveData[22],aCaveData[23]);
        fprintf(theFile,"%.2X %.2X %.2X %.2X         Random objects:\n",aCaveData[24],aCaveData[25],aCaveData[26],aCaveData[27]);
        if (aCaveData[28]) {
            fprintf(theFile,"%.2X %.2X %.2X %.2X             %-7s: %3d/256 = %2d%%\n",
                    aCaveData[28],aCaveData[29],aCaveData[30],aCaveData[31], 
                    GetObjects(aCaveData[24]),aCaveData[28],(int)aCaveData[28]*100/256);
        } else {
            fprintf(theFile,"%.2X %.2X %.2X %.2X             first  code unused (0%%)\n",
                    aCaveData[28],aCaveData[29],aCaveData[30],aCaveData[31]);
        }
        if (aCaveData[29]) {
            fprintf(theFile,"                        %-7s: %3d/256 = %2d%%\n", 
                    GetObjects(aCaveData[25]),aCaveData[29],(int)aCaveData[29]*100/256);
        } else {
            fprintf(theFile,"                        second code unused (0%%)\n");
        }
        if (aCaveData[30]) {
            fprintf(theFile,"                        %-7s: %3d/256 = %2d%%\n", 
                    GetObjects(aCaveData[26]),aCaveData[30],(int)aCaveData[30]*100/256);
        } else {
            fprintf(theFile,"                        third  code unused (0%%)\n");
        }
        if (aCaveData[31]) {
            fprintf(theFile,"                        %-7s: %3d/256 = %2d%%\n", 
                    GetObjects(aCaveData[27]),aCaveData[31],(int)aCaveData[31]*100/256);
        } else {
            fprintf(theFile,"                        fourth code unused (0%%)\n");
        }
        fprintf(theFile,"Cave layout:\n");
    #endif

/* Clear out the cave data to a null value */
    for(x = 0; x < 40; x++) {
        for (y = 0; y <= 23; y++) {
            StoreObject(x, y, 0x07);
        }
    }

/* Decode the random cave data */
    for(y = 3; y <= 23; y++) {
        for(x = 0; x <= 39; x++) {
            theObject = 1;  /* Dirt */
            NextRandom(&RandSeed1, &RandSeed2);
            for (caveDataIndex = 0; caveDataIndex <= 3; caveDataIndex++) {
                if (RandSeed1 < aCaveData[0x1C + caveDataIndex]) {
                    theObject = aCaveData[0x18 + caveDataIndex];
                }
            }
            StoreObject(x, y, theObject);
        }     
    }  

/* SteelBounds */
    DrawRect(0x07, 0, 2, 40, 22);  // before commands!! // x,y,width,height

/* Decode the explicit cave data (commands) */
/* start from "offpara" (usually 0x20)      */
    int featureValue = 0;
    for (caveDataIndex = offpara; aCaveData[caveDataIndex] != 0xFF; caveDataIndex++) 
    {
        int cmd   = 0xFF;
        theCode   = aCaveData[caveDataIndex];
        theObject = theCode & 0x3F;

        cmd = GetCommand(theCode);
        if (GetFeature(c_feature_command0x0F, featureValue) == true) {
            if (featureValue == 1 && theCode == 0x0F) {
                cmd       = theCode;
                theObject = aCaveData[++caveDataIndex];
            } 
        }

        switch(cmd) {   // 3 & (theCode >> 6)
        case 0: /* PLOT */
            x = aCaveData[++caveDataIndex];
            y = aCaveData[++caveDataIndex];
            #if INCLUDE_HUMAN_READABLE
                fprintf(theFile,"%.2X %.2X %.2X            StoreChar %s at (%2d,%2d)\n",theCode,x,y,GetObjects(theCode & 0x3f),x,y);
            #endif
            StoreObject(x, y, theObject);
            break;

        case 1: /* LINE */
            x = aCaveData[++caveDataIndex];
            y = aCaveData[++caveDataIndex];
            theLength = aCaveData[++caveDataIndex];
            theDirection = aCaveData[++caveDataIndex];
            #if INCLUDE_HUMAN_READABLE
                fprintf(theFile,"%.2X %.2X %.2X %.2X %.2X      Line of %s from (%2d,%2d); length = %d; direction = %s\n",theCode,x,y,theLength,theDirection,
                        GetObjects(theCode & 0x3f),x,y,theLength,GetDirections(theDirection));
            #endif
            DrawLine(theObject, x, y, theLength, theDirection);
            break;

        case 2: /* FILLED RECTANGLE */
            x = aCaveData[++caveDataIndex];
            y = aCaveData[++caveDataIndex];
            theWidth  = aCaveData[++caveDataIndex];
            theHeight = aCaveData[++caveDataIndex];
            theFill   = aCaveData[++caveDataIndex];
            #if INCLUDE_HUMAN_READABLE
                fprintf(theFile,"%.2X %.2X %.2X %.2X %.2X %.2X   FilledRect of %s from (%2d,%2d); length = %d; height = %d; fill = %s\n",
                        theCode,x,y,theWidth,theHeight,aCaveData[caveDataIndex],GetObjects(theCode & 0x3f),x,y,theWidth,theHeight,
                        GetObjects(aCaveData[caveDataIndex]));
            #endif
            DrawFilledRect(theObject, x, y, theWidth, theHeight, theFill);
            break;

        case 3: /* OPEN RECTANGLE */
            x = aCaveData[++caveDataIndex];
            y = aCaveData[++caveDataIndex];
            theWidth = aCaveData[++caveDataIndex];
            theHeight = aCaveData[++caveDataIndex];
            #if INCLUDE_HUMAN_READABLE
                fprintf(theFile,"%.2X %.2X %.2X %.2X %.2X      Rect of %s from (%2d,%2d); length = %d; height = %d\n",
                        theCode,x,y,theWidth,theHeight,GetObjects(theCode & 0x3f),x,y,theWidth,theHeight);
            #endif
            DrawRect(theObject, x, y, theWidth, theHeight);
            break;

        case 0x0F: /* dotted */
            x = aCaveData[++caveDataIndex];
            y = aCaveData[++caveDataIndex];
            theXCount = aCaveData[++caveDataIndex];
            theYCount = aCaveData[++caveDataIndex];
            theXSpace = aCaveData[++caveDataIndex];
            theYSpace = aCaveData[++caveDataIndex];
            #if INCLUDE_HUMAN_READABLE
                fprintf(theFile,"%.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X   Dotted of %s from (%2d,%2d); Xcount = %d Ycount = %d; XSpace = %d; YSpace = %d\n",
                        theCode,theObject,x,y,theXCount, theYCount, theXSpace, theYSpace,
                        GetObjects(theObject),x,y,theXCount,theYCount,theXSpace,theYSpace);
            #endif
            for(int iy=0;iy<theYCount;iy++) {
                int xx = x;
                for(int ix=0;ix<theXCount;ix++) {
                    StoreObject(xx, y, theObject);
                    xx = xx + theXSpace;
                    if (xx >= 40) {
                        xx = xx - 40;
                        y  = y + 1;
                    }
                }
                y = y + theYSpace;
            }
            break;
        }

        
    }

/* SteelBounds */
//    DrawRect(0x07, 0, 2, 40, 22);  // before commands!!

/* Print out the resulting aCaveData */
    #if INCLUDE_HUMAN_READABLE
        fprintf(theFile,"FF                  End of cave data.");
    #endif
    fprintf(theFile,"\n\nMap %c #%d\n",caveNo, caveNo-'A'+1);
    for (y = 2; y <= 23; y++) {
        fprintf(theFile, "%03d ", y);
        for(x = 0; x <= 39; x++) {
            fprintf(theFile, "%c", GetCaveData(x, y));  // get them
        }
        if (y == 23) {
            fprintf(theFile, "");
        }
        fprintf(theFile, "\n");
    }
}


/* **************************************** */
void DecodeCaveCommon::SetFeature(std::string key, int value) {
    featureMap[key] = value;
}

/* **************************************** */
bool DecodeCaveCommon::GetFeature(std::string key, int& value) {
    value = 0;
    if (featureMap.find(key) != featureMap.end()) {
        value = featureMap[key];
        return true;
    }
    return false;
}